from __future__ import print_function
from builtins import (range, super)

from . import PropertiesData

from ..functions import RTOL, ATOL


class PropertiesDataBounds(PropertiesData):
    '''Mixin class for a data array with descriptive properties and cell
bounds.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, data=None, bounds=None,
                 geometry=None, interior_ring=None, node_count=None,
                 part_node_count=None, source=None, copy=True,
                 _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
           ``properties={'standard_name': 'longitude'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.
  
    data: `Data`, optional
        Set the data array. Ignored if the *source* parameter is set.
        
        The data array may also be set after initialisation with the
        `set_data` method.
  
    bounds: `Bounds`, optional
        Set the bounds array. Ignored if the *source* parameter is
        set.
        
        The bounds array may also be set after initialisation with the
        `set_bounds` method.
  
    geometry: `str`, optional
        Set the geometry type. Ignored if the *source* parameter is
        set.
        
        *Parameter example:*
           ``geometry='polygon'``
        
        The geometry type may also be set after initialisation with
        the `set_geometry` method.
  
    interior_ring: `InteriorRing`, optional
        Set the interior ring variable. Ignored if the *source*
        parameter is set.
        
        The interior ring variable may also be set after
        initialisation with the `set_interior_ring` method.

    node_count: `NodeCount`, optional
        Set the node count variable for geometry bounds. Ignored if
        the *source* parameter is set.
        
        The node count variable may also be set after initialisation
        with the `set_node_count` method.
    
    part_node_count: `PartNodeCount`, optional
        Set the part node count variable for geometry bounds. Ignored
        if the *source* parameter is set.
        
        The part node count variable may also be set after
        initialisation with the `set_node_count` method.
    
    source: optional
        Initialize the properties, geometry type, data, bounds,
        interior ring variable, node count variable and part node
        count variable from those of *source*.
 
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        # Initialise properties, data, geometry and interior ring
        super().__init__(properties=properties, data=data,
                         bounds=bounds, source=source,
                         geometry=geometry,
                         interior_ring=interior_ring, copy=copy,
                         _use_data=_use_data)

        # Get node count and part node count variables from source
        if source is not None:
            try:
                node_count = source.get_node_count(None)
            except AttributeError:
                node_count = None

            try:
                part_node_count = source.get_part_node_count(None)
            except AttributeError:
                part_node_count = None
        #--- End: if

        # Initialise node count
        if node_count is not None:
            self.set_node_count(node_count, copy=copy)
            
        # Initialise part node count
        if part_node_count is not None:
            self.set_part_node_count(part_node_count, copy=copy)
    #--- End: def

    def __getitem__(self, indices):
        '''Return a subspace of the construct defined by indices

f.__getitem__(indices) <==> f[indices]

The new subspace contains the same properties and similar components
to the original construct, but the latter are also subspaced over
their corresponding axes.

Indexing follows rules that are very similar to the numpy indexing
rules, the only differences being:

* An integer index i takes the i-th element but does not reduce the
  rank by one.

* When two or more dimensions' indices are sequences of integers then
  these indices work independently along each dimension (similar to
  the way vector subscripts work in Fortran). This is the same
  behaviour as indexing on a Variable object of the netCDF4 package.

:Returns:

        The subspace of the construct.

**Examples:**

>>> f.data.shape
(1, 10, 9)
>>> f[:, :, 1].data.shape
(1, 10, 1)
>>> f[:, 0].data.shape
(1, 1, 9)
>>> f[..., 6:3:-1, 3:6].data.shape
(1, 3, 3)
>>> f[0, [2, 9], [4, 8]].data.shape
(1, 2, 2)
>>> f[0, :, -2].data.shape
(1, 10, 1)

        '''
        if indices is Ellipsis: 
            return self.copy() 
 
        if not isinstance(indices, tuple): 
            indices = (indices,) 
 
        new = super().__getitem__(indices) 

        # Subspace the bounds, if there are any. 
        self_bounds = self.get_bounds(None) 
        if self_bounds is not None: 
            data = self_bounds.get_data(None) 
            if data is not None: 
                # There is a bounds array 
                bounds_indices = list(data._parse_indices(indices))
                bounds_indices[-1] = Ellipsis
                if data.ndim <= 1 and not self.has_geometry(): 
                    index = bounds_indices[0] 
                    if isinstance(index, slice): 
                        if index.step < 0: 
                            # This scalar or 1-d variable has been 
                            # reversed so reverse its bounds (as per 
                            # 7.1 of the conventions) 
                            bounds_indices[-1] = slice(None, None, -1)
                    elif data.size > 1 and index[-1] < index[0]: 
                        # This 1-d variable has been reversed so 
                        # reverse its bounds (as per 7.1 of the 
                        # conventions) 
                        bounds_indices[-1] = slice(None, None, -1)
                #--- End: if
 
                new.set_bounds(self_bounds[tuple(bounds_indices)], copy=False) 
        #--- End: if 
 
        # Subspace the interior ring array, if there are one. 
        interior_ring = self.get_interior_ring(None) 
        if interior_ring is not None: 
            new.set_interior_ring(interior_ring[indices], copy=False) 
 
        # Return the new bounded variable 
        return new 
    #--- End: def

    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

.. versionadded:: 1.7.0

        '''
        name = self.identity('')
        
        shape = None        
        data = self.get_data(None)
        bounds = self.get_bounds(None)
        if data is not None:
            shape = data.shape
        else:
            pass
#            bounds_data = bounds.get_data(None)
#            if bounds_data is not None:
#                shape = bounds_data.shape[:-1] # geometry TODO
        #--- End: if
        
        if shape is not None:        
            dims = ', '.join([str(x) for x in shape])
            dims = '({0})'.format(dims)
        else:
            dims = ''
        
#        data = self.get_data(None)
#        if data is not None:
#            dims = ', '.join([str(x) for x in data.shape])
#            dims = '({0})'.format(dims)
#        else:
#            bounds = self.get_bounds(None)
#            if bounds is not None:
#                data = bounds.get_data(None)
#                if data is not None:
#                    dims = ', '.join([str(x) for x in data.shape[:-1]])
#                    dims = '({0})'.format(dims)
#                else:
#                    dims = ''
#            else:
#                dims = ''
#        #--- End: if

        # ------------------------------------------------------------
        # Units and calendar
        # ------------------------------------------------------------
        
        units = self.get_property('units', None)
        if units is None and bounds is not None:
            units = bounds.get_property('units', None)
            
        calendar = self.get_property('calendar', None)
        if calendar is None and bounds is not None:
            calendar = bounds.get_property('calendar', None)

        if units is None:
            isreftime = (calendar is not None)
            units = ''
        else:
            isreftime = 'since' in units
                        
        if isreftime:
            if calendar is None:
                calendar = ''

            units += ' ' + calendar
            
        return '{0}{1} {2}'.format(self.identity(''), dims, units)
    #--- End: def
    
    def del_node_count(self, default=ValueError()):
        '''Remove the node count variable for geometry bounds.

.. versionadded:: 1.8.0

.. seealso:: `get_node_count`, `has_node_count`, `set_node_count`

:Parameters:

    default: optional
        Return the value of the *default* parameter if the node count
        variable has not been set. If set to an `Exception` instance
        then it will be raised instead.

:Returns:

        The removed node count variable.

**Examples:**

>>> n = cfdm.NodeCount(properties={'long_name': 'node counts'})
>>> c.set_node_count(n)
>>> c.has_node_count()
True
>>> c.get_node_count()
<NodeCount: long_name=node counts>
>>> c.del_node_count()
<NodeCount: long_name=node counts>
>>> c.has_node_count()
False

        '''
        try:
            return self._del_component('node_count')
        except ValueError:
            return self._default(default,
                                 "{!r} has no node count variable".format(self.__class__.__name__))
    #--- End: def
    
    def del_part_node_count(self, default=ValueError()):
        '''Remove the part node count variable for geometry bounds.

.. versionadded:: 1.8.0

.. seealso:: `get_part_node_count`, `has_part_node_count`,
             `set_part_node_count`

:Parameters:

    default: optional
        Return the value of the *default* parameter if the part node
        count variable has not been set. If set to an `Exception`
        instance then it will be raised instead.

:Returns:

        The removed part_node count variable.

**Examples:**

>>> p = cfdm.PartNodeCount(properties={'long_name': 'part node counts'})
>>> c.set_part_node_count(p)
>>> c.has_part_node_count()
True
>>> c.get_part_node_count()
<PartNodeCount: long_name=part node counts>
>>> c.del_part_node_count()
<PartNodeCount: long_name=part node counts>
>>> c.has_part_node_count()
False

        '''
        try:
            return self._del_component('part_node_count')
        except ValueError:
            return self._default(default,
                                 "{!r} has no part node count variable".format(self.__class__.__name__))
    #--- End: def

    def dump(self, display=True, _key=None, _omit_properties=None,
             _prefix='', _title=None, _create_title=True, _level=0,
             _axes=None, _axis_names=None):
        '''A full description.

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default

:Returns:

        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        # ------------------------------------------------------------
        # Properties and Data
        # ------------------------------------------------------------
        string = super().dump(display=False, _key=_key,
                              _omit_properties=_omit_properties,
                              _prefix=_prefix, _title=_title,
                              _create_title=_create_title,
                              _level=_level, _axes=_axes,
                              _axis_names=_axis_names)

        string = [string]

        # ------------------------------------------------------------
        # Geometry type
        # ------------------------------------------------------------
        geometry = self.get_geometry(None)
        if geometry is not None:
            indent1 = '    ' * (_level + 1)
            string.append(
                '{0}{1}Geometry: {2}'.format(indent1, _prefix, geometry))

        # ------------------------------------------------------------
        # Bounds
        # ------------------------------------------------------------
        bounds = self.get_bounds(None)
        if bounds is not None:
            string.append(bounds.dump(display=False, _key=_key,
                                      _prefix=_prefix+'Bounds:',
                                      _create_title=False,
                                      _level=_level, _axes=_axes,
                                      _axis_names=_axis_names))

        #-------------------------------------------------------------
        # Interior ring
        # ------------------------------------------------------------
        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            string.append(interior_ring.dump(display=False, _key=_key,
                                             _prefix=_prefix+'Interior Ring:',
                                             _create_title=False,
                                             _level=_level, _axes=_axes,
                                             _axis_names=_axis_names))
            
        string = '\n'.join(string)
        
        if display:
            print(string)
        else:
            return string
    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_properties=(), ignore_compression=False,
               ignore_type=False):
        '''Whether two instances are the same.

Equality is strict by default. This means that:

* the same descriptive properties must be present, with the same
  values and data types, and vector-valued properties must also have
  same the size and be element-wise equal (see the *ignore_properties*
  and *ignore_data_type* parameters), and

..

* if there are data arrays then they must have same shape and data
  type, the same missing data mask, and be element-wise equal (see the
  *ignore_data_type* parameter).

..

* if there are bounds then their descriptive properties (if any) must
  be the same and their data arrays must have same shape and data
  type, the same missing data mask, and be element-wise equal (see the
  *ignore_properties* and *ignore_data_type* parameters).

Two real numbers ``x`` and ``y`` are considered equal if
``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. See the *atol* and *rtol*
parameters.

If data arrays are compressed then the compression type and the
underlying compressed arrays must be the same, as well as the arrays
in their uncompressed forms. See the *ignore_compression* parameter.

Any type of object may be tested but, in general, equality is only
possible with another object of the same type, or a subclass of
one. See the *ignore_type* parameter.

NetCDF elements, such as netCDF variable and dimension names, do not
constitute part of the CF data model and so are not checked.

.. versionadded:: 1.7.0

:Parameters:


    other: 
        The object to compare for equality.

    atol: float, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.
        
    rtol: float, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    ignore_fill_value: `bool`, optional
        If True then the "_FillValue" and "missing_value" properties
        are omitted from the comparison.

    verbose: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_properties: sequence of `str`, optional
        The names of properties to omit from the comparison.

    ignore_data_type: `bool`, optional
        If True then ignore the data types in all numerical
        comparisons. By default different numerical data types imply
        inequality, regardless of whether the elements are within the
        tolerance for equality.

    ignore_compression: `bool`, optional
        If True then any compression applied to the underlying arrays
        is ignored and only the uncompressed arrays are tested for
        equality. By default the compression type and, if appliciable,
        the underlying compressed arrays must be the same, as well as
        the arrays in their uncompressed forms.

    ignore_type: `bool`, optional
        Any type of object may be tested but, in general, equality is
        only possible with another object of the same type, or a
        subclass of one. If *ignore_type* is True then equality is
        possible for any object with a compatible API.

:Returns: 
  
    `bool`
        Whether the two instances are equal.

**Examples:**

>>> p.equals(p)
True
>>> p.equals(p.copy())
True
>>> p.equals('not a colection of properties')
False

        '''    
        # ------------------------------------------------------------
        # Check the properties and data
        # ------------------------------------------------------------
        if not super().equals(other, rtol=rtol, atol=atol,
                              verbose=verbose,
                              ignore_data_type=ignore_data_type,
                              ignore_fill_value=ignore_fill_value,
                              ignore_properties=ignore_properties,
                              ignore_type=ignore_type,
                              ignore_compression=ignore_compression):
            return False
    
        # ------------------------------------------------------------
        # Check the geometry type
        # ------------------------------------------------------------
        if self.get_geometry(None) != other.get_geometry(None):
            if verbose:
                print(
"{0}: Different geometry types: {1}, {2}".format(
    self.__class__.__name__, self.get_geometry(None), other.get_geometry(None)))
            return False

        # ------------------------------------------------------------
        # Check the bounds 
        # ------------------------------------------------------------
        self_has_bounds = self.has_bounds()
        if self_has_bounds != other.has_bounds():
            if verbose:
                print("{0}: Different bounds".format(self.__class__.__name__))
            return False
                
        if self_has_bounds:            
            if not self._equals(self.get_bounds(), other.get_bounds(),
                                rtol=rtol, atol=atol,
                                verbose=verbose,
                                ignore_data_type=ignore_data_type,
                                ignore_type=ignore_type,
                                ignore_fill_value=ignore_fill_value,
                                ignore_compression=ignore_compression):
                if verbose:
                    print("{0}: Different bounds".format(self.__class__.__name__))
                return False
        #--- End: if

        # ------------------------------------------------------------
        # Check the interior ring
        # ------------------------------------------------------------
        self_has_interior_ring = self.has_interior_ring()
        if self_has_interior_ring != other.has_interior_ring():
            if verbose:
                print("{0}: Different interior ring".format(self.__class__.__name__))
            return False
                
        if self_has_interior_ring:            
            if not self._equals(self.get_interior_ring(), other.get_interior_ring(),
                                rtol=rtol, atol=atol,
                                verbose=verbose,
                                ignore_data_type=ignore_data_type,
                                ignore_type=ignore_type,
                                ignore_fill_value=ignore_fill_value,
                                ignore_compression=ignore_compression):
                if verbose:
                    print("{0}: Different interior ring".format(self.__class__.__name__))
                return False
        #--- End: if

        return True
    #--- End: def

    def get_node_count(self, default=ValueError()):
        '''Return the node count variable for geometry bounds.

.. versionadded:: 1.8.0

.. seealso:: del_node_count`, `has_node_count`, `set_node_count`

:Parameters:

    default: optional
        Return the value of the *default* parameter if a node count
        variable has not been set. If set to an `Exception` instance
        then it will be raised instead.

:Returns:

        The node count variable.

**Examples:**

>>> n = cfdm.NodeCount(properties={'long_name': 'node counts'})
>>> c.set_node_count(n)
>>> c.has_node_count()
True
>>> c.get_node_count()
<NodeCount: long_name=node counts>
>>> c.del_node_count()
<NodeCount: long_name=node counts>
>>> c.has_node_count()
False

        '''
        try:
            return self._get_component('node_count')
        except ValueError:
            return self._default(default,
                    "{!r} has no node count variable".format(
                        self.__class__.__name__))
    #--- End: def

    def get_part_node_count(self, default=ValueError()):
        '''Return the part node count variable for geometry bounds.

.. versionadded:: 1.8.0

.. seealso:: `del_part_node_count`, `get_node_count`,
             `has_part_node_count`, `set_part_node_count`

:Parameters:

    default: optional
        Return the value of the *default* parameter if the part node
        count variable has not been set. If set to an `Exception`
        instance then it will be raised instead.

:Returns:

        The part node count variable.

**Examples:**

>>> p = cfdm.PartNodeCount(properties={'long_name': 'part node counts'})
>>> c.set_part_node_count(p)
>>> c.has_part_node_count()
True
>>> c.get_part_node_count()
<PartNodeCount: long_name=part node counts>
>>> c.del_part_node_count()
<PartNodeCount: long_name=part node counts>
>>> c.has_part_node_count()
False

        '''
        try:
            return self._get_component('part_node_count')
        except ValueError:
            return self._default(default,
                    "{!r} has no part node count variable".format(
                        self.__class__.__name__))
    #--- End: def

    def has_node_count(self):
        '''Whether or not there is a node count variable for geometry bounds..
        
.. versionadded:: 1.8.0

.. seealso:: `del_node_count`, `get_node_count`, `set_node_count`

:Returns:

    `bool`
        True if there is a node count variable, otherwise False.

**Examples:**


>>> n = cfdm.NodeCount(properties={'long_name': 'node counts'})
>>> c.set_node_count(n)
>>> c.has_node_count()
True
>>> c.get_node_count()
<NodeCount: long_name=node counts>
>>> c.del_node_count()
<NodeCount: long_name=node counts>
>>> c.has_node_count()
False

        '''
        return self._has_component('node_count')
    #--- End: def

    def has_part_node_count(self):
        '''Whether or not there is a part node count variable for geometry
bounds..
        
.. versionadded:: 1.8.0

.. seealso:: `del_part_node_count`, `get_part_node_count`,
             `set_part_node_count`

:Returns:

    `bool`
        True if there is a part node count variable, otherwise False.

**Examples:**

>>> p = cfdm.PartNodeCount(properties={'long_name': 'part node counts'})
>>> c.set_part_node_count(p)
>>> c.has_part_node_count()
True
>>> c.get_part_node_count()
<PartNodeCount: long_name=part node counts>
>>> c.del_part_node_count()
<PartNodeCount: long_name=part node counts>
>>> c.has_part_node_count()
False

        '''
        return self._has_component('part_node_count')
    #--- End: def

    def identities(self):
        '''Return all possible identities.

The identities comprise:

* The "standard_name" property.
* All properties, preceeded by the property name and a colon,
  e.g. ``'long_name:Air temperature'``.
* The netCDF variable name, preceeded by ``'ncvar%'``.
* The identities of the bounds, if any.

.. versionadded:: 1.7.0

.. seealso:: `identity`

:Returns:

    `list`
        The identities.

**Examples:**

>>> f.properties()
{'foo': 'bar',
 'long_name': 'Air Temperature',
 'standard_name': 'air_temperature'}
>>> f.nc_get_variable()
'tas'
>>> f.identities()
['air_temperature',
 'long_name=Air Temperature',
 'foo=bar',
 'standard_name=air_temperature',
 'ncvar%tas']

>>> f.properties()
{}
>>> f.bounds.properties()
{'axis': 'Z',
 'units': 'm'}
>>> f.identities()
['axis=Z', 'units=m', 'ncvar%z']

        '''
        identities = super().identities()
        
        bounds = self.get_bounds(None)
        if bounds is not None:
            identities.extend([i for i in bounds.identities()
                               if i not in identities])
            
        return identities
    #--- End: def

    def identity(self, default=''):
        '''Return the canonical identity.

By default the identity is the first found of the following:

1. The "standard_name" property.
2. The "cf_role" property, preceeded by ``'cf_role='``.
3. The "axis" property, preceeded by ``'axis='``.
4. The "long_name" property, preceeded by ``'long_name='``.
5. The netCDF variable name, preceeded by ``'ncvar%'``.
6. The identity of the bounds, if any.
7. The value of the *default* parameter.

.. versionadded:: 1.7.0

.. seealso:: `identities`

:Parameters:

    default: optional
        If no identity can be found then return the value of the
        default parameter.

:Returns:

        The identity.

**Examples:**

>>> f.properties()
{'foo': 'bar',
 'long_name': 'Air Temperature',
 'standard_name': 'air_temperature'}
>>> f.nc_get_variable()
'tas'
>>> f.identity()
'air_temperature'
>>> f.del_property('standard_name')
'air_temperature'
>>> f.identity(default='no identity')
'air_temperature'
>>> f.identity()
'long_name=Air Temperature'
>>> f.del_property('long_name')
>>> f.identity()
'ncvar%tas'
>>> f.nc_del_variable()
'tas'
>>> f.identity()
'ncvar%tas'
>>> f.identity()
''
>>> f.identity(default='no identity')
'no identity'

>>> f.properties()
{}
>>> f.bounds.properties()
{'axis': 'Z',
 'units': 'm'}
>>> f.identity()
'axis=Z'

        '''
        identity = super().identity(default=None)
        if identity is not None:
            return identity

        bounds = self.get_bounds(None)
        if bounds is not None:
            return bounds.identity(default=default)
        
        return default
    #--- End: def            
    
    def get_bounds(self, default=ValueError()):
        '''Return the bounds.

.. versionadded:: 1.7.0

.. seealso:: `bounds`, `get_data`, `del_bounds`, `has_bounds`,
             `set_bounds`

:Parameters:

    default: optional
        Return the value of the *default* parameter if bounds have not
        been set. If set to an `Exception` instance then it will be
        raised instead.

:Returns:

        The bounds.

**Examples:**

>>> b = cfdm.Bounds(data=cfdm.Data(range(10).reshape(5, 2)))
>>> c.set_bounds(b)
>>> c.has_bounds()
True
>>> c.get_bounds()
<Bounds: (5, 2) >
>>> b = c.del_bounds()
>>> b
<Bounds: (5, 2) >
>>> c.has_bounds()
False
>>> print(c.get_bounds(None))
None
>>> print(c.del_bounds(None))
None

        '''
        bounds = super().get_bounds(default=None)

        if bounds is None:
            return super().get_bounds(default=default)
        
        inherited_properties = self.properties()
#        bounds_properties = bounds.properties()
#
#        inherited_properties = {prop: value
#                                for prop, value in properties.items()}

        bounds._set_component('inherited_properties', inherited_properties)
        
        return bounds
    #--- End: def

    def get_bounds_data(self, default=ValueError()):
        '''Return the bounds data.

.. versionadded:: 1.7.0

.. seealso:: `bounds`, `get_bounds`, `get_data`

:Parameters:

    default: optional
        Return the value of the *default* parameter if there are no
        bounds data. If set to an `Exception` instance then it will be
        raised instead.

:Returns:

        The bounds data.

**Examples:**

>>> c.get_bounds_data()
<Data(96, 2): [[0, ..., 360.0]] degrees_east>

        '''
        bounds = self.get_bounds(default=None)
        if bounds is None:
            return self.get_bounds(default=default)

        return bounds.get_data(default=default)
    #--- End: def

    def insert_dimension(self, position, inplace=False):
        '''Expand the shape of the data array.

Inserts a new size 1 axis into the data array. A corresponding axis is
also inserted into the bounds data array, if present.

.. versionadded:: 1.7.0

.. seealso:: `squeeze`, `transpose`

:Parameters:

    position: `int`, optional
        Specify the position that the new axis will have in the data
        array. By default the new axis has position 0, the slowest
        varying position. Negative integers counting from the last
        position are allowed.

        *Parameter example:*
          ``position=2``

        *Parameter example:*
          ``position=-1``

    inplace: `bool`, optional
        If True then do the operation in-place and return `None`.

:Returns:

        The new construct with expanded data axes. If the operation
        was in-place then `None` is returned.

**Examples:**

>>> f.data.shape
(19, 73, 96)
>>> f.insert_dimension(position=3).data.shape
(96, 73, 19, 1)
>>> g = f.insert_dimension(position=-1)
>>> g.data.shape
(19, 73, 1, 96)
>>> f.bounds.data.shape
(19, 73, 96, 4)
>>> g.bounds.data.shape
(19, 73, 1, 96, 4)

        '''    
        # Parse position
        ndim = self.data.ndim 
        if -ndim-1 <= position < 0:
            position += ndim + 1
        elif not 0 <= position <= ndim:
            raise ValueError(
                "Can't insert dimension: Invalid position: {!r}".format(position))

#        position = self._parse_axes([position])[0]

        c = super().insert_dimension(position, inplace=inplace)
        if inplace:
            c = self
        
        # ------------------------------------------------------------
        # Expand the dims of the bounds
        # ------------------------------------------------------------
        bounds = c.get_bounds(None)
        if bounds is not None:
            bounds.insert_dimension(position, inplace=True)

        # ------------------------------------------------------------
        # Expand the dims of the interior_ring
        # ------------------------------------------------------------
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring.insert_dimension(position, inplace=True)

        if inplace:
            c = None
            
        return c
    #--- End: def

    def set_node_count(self, node_count, copy=True):
        '''Set the node count variable for geometry bounds.

.. versionadded:: 1.8.0

.. seealso: `del_node_count`, `get_node_count`, `has_node_count`

:Parameters:

    node_count: `NodeCount`
        The node count variable to be inserted.

    copy: `bool`, optional
        If False then do not copy the node count variable prior to
        insertion. By default it is copied.

:Returns:

    `None`

**Examples:**

TODO

        '''
        if copy:
            node_count = node_count.copy()

        self._set_component('node_count', node_count, copy=False)
    #--- End: def

    def set_part_node_count(self, part_node_count, copy=True):
        '''Set the part node count variable for geometry bounds.

.. versionadded:: 1.8.0

.. seealso: `del_part_node_count`, `get_part_node_count`,
            `has_part_node_count`

:Parameters:

    part_node_count: `PartNodeCount`
        The part node count variable to be inserted.

    copy: `bool`, optional
        If False then do not copy the part node count variable prior
        to insertion. By default it is copied.

:Returns:

    `None`

**Examples:**

TODO

        '''
        if copy:
            part_node_count = part_node_count.copy()

        self._set_component('part_node_count', part_node_count, copy=False)
    #--- End: def

    def squeeze(self, axes=None, inplace=False):
        '''Remove size one axes from the data array.

By default all size one axes are removed, but particular size one axes
may be selected for removal. Corresponding axes are also removed from
the bounds data array, if present.

.. versionadded:: 1.7.0

.. seealso:: `insert_dimension`, `transpose`

:Parameters:

    axes: (sequence of) `int`
        The positions of the size one axes to be removed. By default
        all size one axes are removed. Each axis is identified by its
        original integer position. Negative integers counting from the
        last position are allowed.

        *Parameter example:*
          ``axes=0``

        *Parameter example:*
          ``axes=-2``

        *Parameter example:*
          ``axes=[2, 0]``

    inplace: `bool`, optional
        If True then do the operation in-place and return `None`.

:Returns:

        The new construct with removed data axes. If the operation was
        in-place then `None` is returned.

**Examples:**

>>> f.data.shape
(1, 73, 1, 96)
>>> f.squeeze().data.shape
(73, 96)
>>> f.squeeze(0).data.shape
(73, 1, 96)
>>> g = f.squeeze([-3, 2])
>>> g.data.shape
(73, 96)
>>> f.bounds.data.shape
(1, 73, 1, 96, 4)
>>> g.data.shape
(73, 96, 4)

        '''     
        axes = self._parse_axes(axes)

        c = super().squeeze(axes, inplace=inplace)
        if inplace:
            c = self
        
        # ------------------------------------------------------------
        # Squeeze the bounds
        # ------------------------------------------------------------
        bounds = c.get_bounds(None)
        if bounds is not None:
            bounds.squeeze(axes, inplace=True)

        # ------------------------------------------------------------
        # Squeeze the interior_ring
        # ------------------------------------------------------------
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring.squeeze(axes, inplace=True)

        if inplace:
            c = None
                
        return c
    #--- End: def
    
    def transpose(self, axes=None, inplace=False):
        '''Permute the axes of the data array.

Corresponding axes of the bounds data array, if present, are also
permuted.

Note that if i) the data array is two-dimensional, ii) the two axes
have been permuted, and iii) each cell has four bounds values; then
columns 1 and 3 (counting from 0) of the bounds axis are swapped to
preserve contiguity bounds in adjacent cells. See section 7.1 "Cell
Boundaries" of the CF conventions for details.

.. seealso:: `insert_dimension`, `squeeze`

:Parameters:

    axes: (sequence of) `int`
        The new axis order. By default the order is reversed. Each
        axis in the new order is identified by its original integer
        position. Negative integers counting from the last position
        are allowed.

        *Parameter example:*
          ``axes=[2, 0, 1]``

        *Parameter example:*
          ``axes=[-1, 0, 1]``

    inplace: `bool`, optional
        If True then do the operation in-place and return `None`.

:Returns:

        The new construct with permuted data axes. If the operation
        was in-place then `None` is returned.

**Examples:**

>>> f.data.shape
(19, 73, 96)
>>> f.tranpose().data.shape
(96, 73, 19)
>>> g = f.tranpose([1, 0, 2])
>>> g.data.shape
(73, 19, 96)
>>> f.bounds.data.shape
(19, 73, 96, 4)
>>> g.bounds.data.shape
(73, 19, 96, 4)

        '''
        ndim = self.data.ndim 
        if axes is None: 
            axes = list(range(ndim-1, -1, -1)) 
        else: 
            axes = self._parse_axes(axes) 

        # ------------------------------------------------------------ 
        # Transpose the coordinates 
        # ------------------------------------------------------------         
        c = super().transpose(axes, inplace=inplace)
        if inplace:
            c = self
 
        # ------------------------------------------------------------ 
        # Transpose the bounds 
        # ------------------------------------------------------------         
        bounds = c.get_bounds(None) 
        data = c.get_bounds_data(None) 
        if bounds is not None: 
            b_axes = axes[:] 
            b_axes.extend(list(range(ndim, data.ndim))) 
            bounds.transpose(b_axes, inplace=True) 

            data = bounds.get_data(None)
            if data is not None:
                if (ndim == 2 and data.ndim == 3 and data.shape[-1] == 4 and 
                    b_axes[0:2] == [1, 0]):
#                    print ('UUUUUUUUUUUUUUUUU')
                    # Swap elements 1 and 3 of the trailing dimension so 
                    # that the values are still contiguous (if they ever 
                    # were). See section 7.1 of the CF conventions. 
#                    print (data[2 ,3, slice(1, 4, 2)].array)
#                    print (data[2, 3, slice(3, 0, -2)].array)
                    data[:, :, slice(1, 4, 2)] = data[:, :, slice(3, 0, -2)]
                    bounds.set_data(data, copy=False) 
        #--- End: if 
 
        # ------------------------------------------------------------ 
        # Transpose the interior ring 
        # ------------------------------------------------------------ 
        interior_ring = c.get_interior_ring(None) 
        if interior_ring is not None: 
            interior_ring.transpose(axes + [-1], inplace=True)

        if inplace:
            c = None
            
        return c
    #--- End: def
    
#--- End: class



from __future__ import print_function
from builtins import range

import copy
import datetime
import inspect
import itertools
from operator import mul
import os
import sys
import time
import unittest


import numpy

import cfdm

class DataTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    test_only = []
#    test_only = ['NOTHING!!!!!']
#    test_only = ['test_Data__asdatetime__asreftime__isdatetime']
#    test_only = ['test_Data__setitem__']
#    test_only = ['test_Data_ceil', 'test_Data_floor', 'test_Data_trunc', 'test_Data_rint']
#    test_only = ['test_Data_array', 'test_Data_datetime_array']
#    test_only = ['test_dumpd_loadd']
#    test_only = ['test_Data_BINARY_AND_UNARY_OPERATORS']

    def test_Data__repr__str(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for d in [cfdm.Data(9, units='km'),
                  cfdm.Data([9], units='km'),
                  cfdm.Data([[9]], units='km'),
                  cfdm.Data([8, 9], units='km'),
                  cfdm.Data([[8, 9]], units='km'),
                  cfdm.Data([7, 8, 9], units='km'),
                  cfdm.Data([[7, 8, 9]], units='km'),
                  cfdm.Data([6, 7, 8, 9], units='km'),
                  cfdm.Data([[6, 7, 8, 9]], units='km'),
                  cfdm.Data([[6, 7], [8, 9]], units='km'),
                  cfdm.Data([[6, 7, 8, 9], [6, 7, 8, 9]], units='km'),
        ]:
            _ = repr(d)
            _ = str(d)
    #--- End: def

#    def test_Data__getitem__(self):
#        if self.test_only and inspect.stack()[0][3] not in self.test_only:
#            return
                

    def test_Data__setitem__(self):        
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        a = numpy.ma.arange(3000).reshape(50, 60)

        d = cfdm.Data(a.filled(), units='m')
        
        for n, (j, i) in enumerate(((34, 23), (0, 0), (-1, -1),
                                    (slice(40, 50), slice(58, 60)),
                                    (Ellipsis, slice(None)),
                                    (slice(None), Ellipsis),
                                  )):
            n = -n-1
            for dvalue, avalue in ((n, n), (cfdm.masked, numpy.ma.masked), (n, n)):
                message = "cfdm.Data[{}, {}]={}={} failed".format(j, i, dvalue, avalue)
                d[j, i] = dvalue
                a[j, i] = avalue
                x = d.array
                self.assertTrue((x == a).all() in (True, numpy.ma.masked), message)
                m = numpy.ma.getmaskarray(x)
                self.assertTrue((m == numpy.ma.getmaskarray(a)).all(), 
                                'd.mask.array='+repr(m)+'\nnumpy.ma.getmaskarray(a)='+repr(numpy.ma.getmaskarray(a)))
        #--- End: for
    
        a = numpy.ma.arange(3000).reshape(50, 60)
        
        d = cfdm.Data(a.filled(), 'm')
        
        (j, i) = (slice(0, 2), slice(0, 3))
        array = numpy.array([[1, 2, 6],[3, 4, 5]])*-1
        
        for dvalue in (array,
                       numpy.ma.masked_where(array < -2, array),
                       array):
            message = 'cfdm.Data[%s, %s]=%s failed' % (j, i, dvalue)
            d[j, i] = dvalue
            a[j, i] = dvalue
            x = d.array
            self.assertTrue((x == a).all() in (True, numpy.ma.masked), message)
            m = numpy.ma.getmaskarray(x)
            self.assertTrue((m == numpy.ma.getmaskarray(a)).all(), message)
        #--- End: for
        
        # Scalar numeric array
        d = cfdm.Data(9, units='km')
        d[...] = cfdm.masked
        a = d.array
        self.assertTrue(a.shape == ())
        self.assertTrue(a[()] is numpy.ma.masked)
    #--- End: def

#    def test_Data_astype(self):
#        if self.test_only and inspect.stack()[0][3] not in self.test_only:
#            return
#
#        a = numpy.array([1.5, 2, 2.5], dtype=float)       
#        d = cfdm.Data(a)
#        
#        self.assertTrue(d.dtype == numpy.dtype(float))
#        self.assertTrue(d.array.dtype == numpy.dtype(float))
#        self.assertTrue((d.array == a).all())
#
#        d.astype('int32')
#        self.assertTrue(d.dtype == numpy.dtype('int32'))
#        self.assertTrue(d.array.dtype == numpy.dtype('int32'))
#        self.assertTrue((d.array == [1, 2, 2]).all())
#
#        d = cfdm.Data(a)
#        try:
#            d.astype(numpy.dtype(int, casting='safe'))
#            self.assertTrue(False)
#        except TypeError:
#            pass
#    #--- End: def

    def test_Data_array(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # ------------------------------------------------------------
        # Numpy array interface (__array__)
        # ------------------------------------------------------------
        a = numpy.arange(12, dtype='int32').reshape(3, 4)

        d = cfdm.Data(a, units='km')

        b = numpy.array(d)

        self.assertTrue(b.dtype == numpy.dtype('int32'))
        self.assertTrue(a.shape == b.shape)
        self.assertTrue((a == b).all())        

        b = numpy.array(d, dtype='float32')

        self.assertTrue(b.dtype == numpy.dtype('float32'))
        self.assertTrue(a.shape == b.shape)
        self.assertTrue((a == b).all())

        
        # Scalar numeric array
        d = cfdm.Data(9, units='km')
        a = d.array
        self.assertTrue(a.shape == ())
        self.assertTrue(a == numpy.array(9))
        d[...] = cfdm.masked
        a = d.array
        self.assertTrue(a.shape == ())
        self.assertTrue(a[()] is numpy.ma.masked)

        # Non-scalar numeric array
        b = numpy.arange(10*15*19).reshape(10, 1, 15, 19)
        d = cfdm.Data(b, 'km')
        a = d.array
        a[0,0,0,0] = -999
        a2 = d.array
        self.assertTrue(a2[0,0,0,0] == 0)
        self.assertTrue(a2.shape == b.shape)
        self.assertTrue((a2 == b).all())
        self.assertFalse((a2 == a).all())
    #--- End: def

    def test_Data_datetime_array(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        d = cfdm.Data([11292.5, 11293], units='days since 1970-1-1')
        dt = d.datetime_array
        self.assertTrue(dt[0] == datetime.datetime(2000, 12, 1, 12, 0))
        self.assertTrue(dt[1] == datetime.datetime(2000, 12, 2,  0, 0))

        d[0] = cfdm.masked
        dt = d.datetime_array
        self.assertTrue(dt[0] is numpy.ma.masked)
        self.assertTrue(dt[1] == datetime.datetime(2000, 12, 2,  0, 0))
        
        d = cfdm.Data(11292.5, units='days since 1970-1-1')
        dt = d.datetime_array
        self.assertTrue(dt[()] == datetime.datetime(2000, 12, 1, 12, 0))

        d[()] = cfdm.masked
        dt = d.datetime_array
        self.assertTrue(dt[()] is numpy.ma.masked)
    #--- End: def

    def test_Data_transpose(self):        
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        a = numpy.arange(10*15*19).reshape(10, 1, 15, 19)
        d = cfdm.Data(a.copy())
        
        for indices in (list(range(a.ndim)), list(range(-a.ndim, 0))):
            for axes in itertools.permutations(indices):
                a = numpy.transpose(a, axes)
                d = d.transpose(axes)
                message = 'cfdm.Data.transpose({}) failed: d.shape={}, a.shape={}'.format(
                    axes, d.shape, a.shape)
                self.assertTrue(d.shape == a.shape, message)
                self.assertTrue((d.array == a).all(), message)
            #--- End: for
        #--- End: for
    #--- End: def
        
    def test_Data_unique(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
            
        d = cfdm.Data([[4, 2, 1], [1, 2, 3]], units='metre')
        u = d.unique()
        self.assertTrue(u.shape == (4,))
        self.assertTrue((u.array == cfdm.Data([1, 2, 3, 4], 'metre').array).all())
        

        d[1, -1] = cfdm.masked
        u = d.unique()
        self.assertTrue(u.shape == (3,))        
        self.assertTrue((u.array == cfdm.Data([1, 2, 4], 'metre').array).all())
    #--- End: def

    def test_Data_equals(self):        
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        a = numpy.ma.arange(10*15*19).reshape(10, 1, 15, 19)
        a[0, 0, 2, 3] = numpy.ma.masked
        
        d = cfdm.Data(a, units='days since 2000-2-2', calendar='noleap')
        e = copy.deepcopy(d)

        self.assertTrue(d.equals(d, verbose=True))
        self.assertTrue(d.equals(e, verbose=True))
        self.assertTrue(e.equals(d, verbose=True))    
    #--- End: def
        
    def test_Data_max_min_sum_squeeze(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        
        a = numpy.ma.arange(10*15*19).reshape(10, 1, 15, 19)
        a[0, 0, 0, 0] = numpy.ma.masked
        a[-1, -1, -1, -1] = numpy.ma.masked
        a[5, -1, 6, 10] = numpy.ma.masked
        d = cfdm.Data(a)

        b = a.max()
        x = d.max().squeeze()
        self.assertTrue(x.shape == b.shape)
        self.assertTrue((x.array == b).all())

        b = a.max(axis=0)
        x = d.max(axes=0).squeeze(0)
        self.assertTrue(x.shape == b.shape)
        self.assertTrue((x.array == b).all())

        b = a.max(axis=(0, 3))
        x = d.max(axes=[0, 3]).squeeze([0, 3])
        self.assertTrue(x.shape == b.shape)
        self.assertTrue((x.array == b).all())

        b = a.min()
        x = d.min().squeeze()
        self.assertTrue(x.shape == b.shape)
        self.assertTrue((x.array == b).all())

        b = a.min(axis=(0, 3))
        x = d.min(axes=[0, 3]).squeeze([0, 3])
        self.assertTrue(x.shape == b.shape)
        self.assertTrue((x.array == b).all(), (x.shape, b.shape))

        b = a.sum()
        x = d.sum().squeeze()
        self.assertTrue(x.shape== b.shape)
        self.assertTrue((x.array == b).all())

        b = a.sum(axis=(0, 3))
        x = d.sum(axes=[0, 3]).squeeze([0, 3])
        self.assertTrue(x.shape == b.shape)
        self.assertTrue((x.array == b).all(), (x.shape, b.shape))
    #--- End: def

#--- End: class


if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from builtins import super

from . import mixin
from . import core


class PartNodeCountProperties(mixin.NetCDFVariable,
                              mixin.NetCDFDimension,
                              mixin.Properties,
                              core.abstract.Properties):
    '''Properties for a netCDF part node count variable.

**NetCDF interface**

The netCDF part node count variable name may be accessed with the
`nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
`nc_has_variable` methods.

The name of the netCDF dimension spanned by the netCDF part node count
variable's data may be accessed with the `nc_set_dimension`,
`nc_get_dimension`, `nc_del_dimension` and `nc_has_dimension` methods.

.. versionadded:: 1.8.0

    '''
    def __init__(self, properties=None, source=None, copy=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'long_name': 'number of obs for this station'}``

        Properties may also be set after initialisation with the
        `set_properties` and `set_property` methods.

    source: optional
        Initialize the properties from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, source=source,
                         copy=copy)

        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, _key=None, _title=None,
             _create_title=True, _prefix='', _level=0,
             _omit_properties=None):
        '''A full description of the part node count variable.

Returns a description of all properties.

.. versionadded:: 1.8.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        if _create_title and _title is None: 
            _title = 'Part Node Count: ' + self.identity(default='')

        return super().dump(display=display, _key=_key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title)
    #--- End: def
    
#--- End: class



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str

import os
import sys

from platform import platform, python_version

import netCDF4
#import cftime
import numpy
import future

from . import (__version__,
               __cf_version__,
               __file__)

from .constants import CONSTANTS


def ATOL(*atol):
    '''The tolerance on absolute differences when testing for numerically
tolerant equality.

Two real numbers ``x`` and ``y`` are considered equal if ``abs(x-y) <=
atol + rtol*abs(y)``, where atol (the tolerance on absolute
differences) and rtol (the tolerance on relative differences) are
positive, typically very small numbers. By default both are set to the
system epsilon (the difference between 1 and the least value greater
than 1 that is representable as a float).

.. versionadded:: 1.7.0

.. seealso:: `RTOL`

:Parameters:

    atol: `float`, optional
        The new value of absolute tolerance. The default is to not
        change the current value.

:Returns:

    `float`
        The value prior to the change, or the current value if no
        new value was specified.

**Examples:**

>>> ATOL()
2.220446049250313e-16
>>> old = ATOL(1e-10)
>>> ATOL(old)
1e-10
>>> ATOL()
2.220446049250313e-16

    '''
    old = CONSTANTS['ATOL']
    if atol:
        CONSTANTS['ATOL'] = atol[0]

    return old
#--- End: def

def RTOL(*rtol):    
    '''The tolerance on relative differences when testing for numerically
tolerant equality.

Two real numbers ``x`` and ``y`` are considered equal if ``abs(x-y) <=
atol + rtol*abs(y)``, where atol (the tolerance on absolute
differences) and rtol (the tolerance on relative differences) are
positive, typically very small numbers. By default both are set to the
system epsilon (the difference between 1 and the least value greater
than 1 that is representable as a float).

.. versionadded:: 1.7.0

.. seealso:: `ATOL`

:Parameters:

    rtol: `float`, optional
        The new value of relative tolerance. The default is to not
        change the current value.

:Returns:

    `float`
        The value prior to the change, or the current value if no
        new value was specified.

**Examples:**

>>> RTOL()
2.220446049250313e-16
>>> old = RTOL(1e-10)
>>> RTOL(old)
1e-10
>>> RTOL()
2.220446049250313e-16

    '''
    old = CONSTANTS['RTOL']
    if rtol:
        CONSTANTS['RTOL'] = rtol[0]

    return old
#--- End: def

def environment(display=True):
    '''Return the names, versions and paths of all dependencies.

.. versionadded:: 1.7.0

:Parameters:

    display: `bool`, optional
        If False then return the description of the environment as a
        string. By default the description is printed.

:Returns:

    `None` or `str`
        If *display* is True then the description of the environment
        is printed and `None` is returned. Otherwise the description
        is returned as a string.

**Examples:**

>>> environment()
Platform: Linux-4.15.0-43-generic-x86_64-with-debian-stretch-sid
python: 3.6.5 /home/david/anaconda3/bin/python
HDF5 library: 1.10.2
netcdf library: 4.6.1
netCDF4: 1.4.0 /home/user/anaconda3/lib/python3.6/site-packages/netCDF4/__init__.py
numpy: 1.14.3 /home/user/anaconda3/lib/python3.6/site-packages/numpy/__init__.py
future: 0.16.0 /home/user/anaconda3/lib/python3.6/site-packages/future/__init__.py
cfdm: 1.7.0 /home/user/cfdm/cfdm/__init__.py

    '''
    out = []
    out.append('Platform: ' + str(platform()))
    out.append('python: ' + str(python_version() + ' ' + str(sys.executable)))

    out.append('HDF5 library: ' + str(netCDF4. __hdf5libversion__))
    out.append('netcdf library: ' + str(netCDF4.__netcdf4libversion__))

    out.append('netCDF4: ' + str(netCDF4.__version__) + ' ' + str(os.path.abspath(netCDF4.__file__)))
#    out.append('cftime: ' + str(cftime.__version__) + ' ' + str(os.path.abspath(cftime.__file__)))
    out.append('numpy: ' + str(numpy.__version__) + ' ' + str(os.path.abspath(numpy.__file__)))
    out.append('future: ' + str(future.__version__) + ' ' + str(os.path.abspath(future.__file__)))

    out.append('cfdm: ' + str(__version__) + ' ' + str(os.path.abspath(__file__)))
    
    out = '\n'.join(out)

    if display:
        print(out)
    else:
        return out
#--- End: def

def CF():
    '''The version of the CF conventions.

This indicates which version of the CF conventions are represented by
this release of the cfdm package, and therefore the version can not be
changed.

.. versionadded:: 1.7.0

:Returns:

    `str`
        The version of the CF conventions represented by this release
        of the cfdm package.

**Examples:**

>>> CF()
'1.7'

    '''
    return __cf_version__
#--- End: def




from builtins import (str, super)
from past.builtins import basestring

from copy import deepcopy

from . import abstract


#class CellMethod(abstract.Properties):
class CellMethod(abstract.Container):
    '''A cell method construct of the CF data model.

    One or more cell method constructs describe how the cell values of
    the field construct represent the variation of the physical
    quantity within its cells, i.e. the structure of the data at a
    higher resolution.
    
    A single cell method construct consists of a set of axes, a
    "method" property which describes how a value of the field
    construct's data array describes the variation of the quantity
    within a cell over those axes (e.g. a value might represent the
    cell area average), and descriptive qualifiers serving to indicate
    more precisely how the method was applied (e.g. recording the
    spacing of the original data, or the fact that the method was
    applied only over El Nino years).
    
    .. versionadded:: 1.7.0

    '''
    def __init__(self, axes=None, method=None, qualifiers=None,
                 source=None, copy=True):
        '''**Initialisation**

    :Parameters:
    
        axes: (sequence of) `str`, optional
            Set the axes of the cell method construct, specified
            either by the construct identifiers of domain axis
            constructs, standard names, or the special string
            ``'area'``.
    
            The axes may also be set after initialisation with the
            `set_axes` method.
   
            *Parameter example:*
              ``axes='domainaxis0'``
            
            *Parameter example:*
              ``axes=['domainaxis0']``
            
            *Parameter example:*
              ``axes=('domainaxis0', 'domainaxis1')``
            
            *Parameter example:*
              ``axes='area'``
            
            *Parameter example:*
              ``axes=['domainaxis2', 'time']``
            
        method: `str`, optional
            Set the axes of the cell method construct. Either one or
            more domain axis construct identifiers or standard
            names. Ignored if the *source* parameter is set.
    
            The method may also be set after initialisation with the
            `set_method` method.
    
            *Parameter example:*
              ``method='mean'
    
        qualifiers: `dict`, optional
            Set descriptive qualifiers. The dictionary keys are
            qualifer names, with corresponding values. Ignored if the
            *source* parameter is set.
    
            Qualifiers may also be set after initialisation with the
            `qualifiers` and `set_qualifier` methods.
    
            *Parameter example:*
              ``qualifiers={'comment': 'sampled instantaneously'}``
           
            *Parameter example:*
              ``qualifiers={'where': 'sea', ''over': 'ice'}``
            
        source: optional
            Initialize the axes, method and qualifiers from those of
            *source*.
    
        copy: `bool`, optional
            If False then do not deep copy input parameters prior to
            initialization By default parameters are deep copied.

        '''
        super().__init__()

        if source:
            try:
                axes = source.get_axes(None)
            except AttributeErrror:
                axes = None              

            try:
                method = source.get_method(None)
            except AttributeErrror:
                method = None              

            try:
                qualifiers = source.qualifiers()
            except AttributeError:
                qualifiers = None
        #--- End: if

        if axes is not None:                
            axes = self.set_axes(axes)

        if method is not None:                
            method = self.set_method(method)

        if qualifiers is None:
            qualifiers = {}

        self._set_component('qualifiers', qualifiers, copy=copy)


    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
    .. versionadded:: 1.7.0
            
    :Returns:
    
        `str`
            The construct type.
    
    **Examples:**
    
    >>> f.construct_type
    'cell_method'

        '''
        return 'cell_method'


    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self):
        '''Return a deep copy.

    ``f.copy()`` is equivalent to ``copy.deepcopy(f)``.
    
    .. versionadded:: 1.7.0
    
    :Returns:
    
            The deep copy.
    
    **Examples:**
    
    >>> g = f.copy()

        '''
        return type(self)(source=self, copy=True)


    def del_axes(self, default=ValueError()):
        '''Remove the axes of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `get_axes`, `has_axes`, `set_axes`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if axes have
            not been set. If set to an `Exception` instance then it
            will be raised instead.
    
    :Returns:
    
        `tuple`
            The removed axes, identified by domain axis construct key
            or standard name.
    
    **Examples:**
    
    >>> c.set_axes('domainaxis1')
    >>> c.has_axes()
    True
    >>> c.get_axes()
    ('domainaxis1',)
    >>> c.del_axes()
    >>> c.has_axes()
    False
    >>> c.get_axes('NO AXES')
    'NO AXES'
    >>> c.del_axes('NO AXES')
    'NO AXES'

        '''
        try:
            return self._del_component('axes')
        except ValueError:
            return self._default(default,
                                 "{!r} has no axes".format(
                                     self.__class__.__name__))

    
    def del_method(self, default=ValueError()):
        '''Remove the method of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `get_method`, `has_method`, `set_method`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if the method
            has not been set. If set to an `Exception` instance then
            it will be raised instead.
    
    :Returns:
    
        `str`
            The removed method.
    
    **Examples:**
    
    >>> c.set_method('minimum')
    >>> c.has_method()
    True
    >>> c.get_method()
    ('time',)
    >>> c.del_method()
    >>> c.has_method()
    False
    >>> c.get_method('NO METHOD')
    'NO METHOD'
    >>> c.del_method('NO METHOD')
    'NO METHOD'

        '''
        try:
            return self._del_component('method')
        except ValueError:
            return self._default(default,
                                 "{!r} has no method".format(
                                     self.__class__.__name__))

    
    def del_qualifier(self, qualifier, default=ValueError()):
        '''Remove a qualifier of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `get_qualifier`, `has_qualifier`, `qualifiers`,
                 `set_qualifier`
    
    :Parameters:
    
        qualifier:
            The name of the qualifier to be removed.
    
            *Parameter example:*
              ``qualifier='where'``
    
        default: optional
            Return the value of the *default* parameter if the
            qualifier has not been set. If set to an `Exception`
            instance then it will be raised instead.
    
    :Returns:
    
            The removed qualifier.
    
    **Examples:**
    
    >>> c.set_qualifier('where', 'land')
    >>> c.get_qualifier('where', 'no qualifier')
    'land'
    >>> c.del_qualifier('where')
    'land'
    >>> c.get_qualifier('where')
    ValueError: 'CellMethod' has no 'where' qualifier
    >>> c.del_qualifier('where', 'no qualifier')
    'no qualifier'

        '''
        try:
            return self._get_component('qualifiers').pop(qualifier)
        except KeyError:
            return self._default(default, "{!r} has no {!r} qualifier".format(
                self.__class__.__name__, qualifier))


    def get_axes(self, default=ValueError()):
        '''Return the axes of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_axes`, `has_axes`, `set_axes`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if axes have
            not been set. If set to an `Exception` instance then it
            will be raised instead.
    
    :Returns:
    
        `tuple`
            The axes.
    
    **Examples:**
    
    >>> c.set_axes('domainaxis1')
    >>> c.has_axes()
    True
    >>> c.get_axes()
    ('domainaxis1',)
    >>> c.del_axes()
    >>> c.has_axes()
    False
    >>> c.get_axes('NO AXES')
    'NO AXES'
    >>> c.del_axes('NO AXES')
    'NO AXES'

        '''
        try:
            return self._get_component('axes')
        except ValueError:
            return self._default(default,
                                 "{!r} has no axes".format(
                                     self.__class__.__name__))


    def get_method(self, default=ValueError()):
        '''Return the method of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_method`, `has_method`, `set_method`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if the method
            has not been set. If set to an `Exception` instance then
            it will be raised instead.
    
    :Returns:
    
        `str`
            The method.
    
    **Examples:**
    
    >>> c.set_method('minimum')
    >>> c.has_method()
    True
    >>> c.get_method()
    'minimum'
    >>> c.del_method()
    >>> c.has_method()
    False
    >>> c.get_method('NO METHOD')
    'NO METHOD'
    >>> c.del_method('NO METHOD')
    'NO METHOD'

        '''
        try:
            return self._get_component('method')
        except ValueError:
            return self._default(default,
                                 "{!r} has no method".format(
                                     self.__class__.__name__))


    def get_qualifier(self, qualifier, default=ValueError()):
        '''Return a qualifier of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_qualifier`, `has_qualifier`, `qualifiers`,
                 `set_qualifier`
    
    :Parameters:
    
        qualifier:
            The name of the qualifier to be returned.
    
            *Parameter example:*
              ``qualifier='where'``
    
        default: optional
            Return the value of the *default* parameter if the
            qualifier has not been set. If set to an `Exception`
            instance then it will be raised instead.
    
    :Returns:
    
            The value of the qualifier. 
    
    **Examples:**
    
    >>> c.set_qualifier('where', 'land')
    >>> c.get_qualifier('where', 'no qualifier')
    'land'
    >>> c.del_qualifier('where')
    'land'
    >>> c.get_qualifier('where')
    ValueError: 'CellMethod' has no 'where' qualifier
    >>> c.get_qualifier('where', 'no qualifier')
    'no qualifier'

        '''
        try:
            return self._get_component('qualifiers')[qualifier]
        except KeyError:
            return self._default(default, "{!r} has no {!r} qualifier".format(
                self.__class__.__name__, qualifier))


    def has_axes(self):
        '''Whether the axes of the cell method have been set.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_axes`, `get_axes`, `set_axes`
    
    :Returns:
    
        `bool`
            `True` if the axes have been set, otherwise `False`.
    
    **Examples:**
    
    >>> c.set_axes('domainaxis1')
    >>> c.has_axes()
    True
    >>> c.get_axes()
    ('domainaxis1',)
    >>> c.del_axes()
    >>> c.has_axes()
    False
    >>> c.get_axes('NO AXES')
    'NO AXES'
    >>> c.del_axes('NO AXES')
    'NO AXES'

        '''
        return self._has_component('axes')


    def has_method(self):
        '''Whether the method of the cell method has been set.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_axes`, `get_method`, `set_method`
    
    :Returns:
    
        `bool`
            `True` if the method has been set, otherwise `False`.
    
    **Examples:**
    
    >>> c.set_method('minimum')
    >>> c.has_method()
    True
    >>> c.get_method()
    'minimum'
    >>> c.del_method()
    >>> c.has_method()
    False
    >>> c.get_method('NO METHOD')
    'NO METHOD'
    >>> c.del_method('NO METHOD')
    'NO METHOD'

        '''
        return self._has_component('method')


    def has_qualifier(self, qualifier):
        '''Whether a qualifier of the cell method has been set.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_qualifier`, `get_qualifier`, `qualifiers`,
                 `set_qualifier`
    
    :Parameters:
    
        qualifier: `str`
            The name of the qualifier.
    
    :Returns:
    
        `bool`
            `True` if the qualifier has been set, otherwise `False`.
    
    **Examples:**
    
    >>> c.set_qualifier('where', 'land')
    >>> c.has_qualifier('where')
    True
    >>> c.del_qualifier('where')
    'land'
    >>> c.has_qualifier('where')
    False

        '''
        return qualifier in self._get_component('qualifiers')


    def qualifiers(self):
        '''Return all qualifiers of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_qualifier`, `get_qualifier`, `has_qualifier`,
                 `set_qualifier`
    
    :Returns:
    
        `dict`
            The qualifiers.
    
    **Examples:**
    
    >>> c.qualifiers()
    {'interval': [<Data(): 0.1 degrees>], 
     'where': 'land'}
    
    >>> f.qualifiers()
    {}

        '''
        return self._get_component('qualifiers').copy()


    def set_axes(self, value, copy=True):
        '''Set the axes of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_axes`, `get_axes`, `has_axes`
    
    :Parameters:
    
        value: (sequence of) `str`
            The axes, specified either by the construct identifiers of
            domain axis constructs, standard names, or the special
            string ``'area'``.
    
            *Parameter example:*
              ``axes='domainaxis0'``
    
            *Parameter example:*
              ``axes='time'``
    
            *Parameter example:*
              ``axes='area'``
    
            *Parameter example:*
              ``axes=['domainaxis0', 'domainaxis2']``
    
            *Parameter example:*
              ``axes=['time', 'area']``
    
            *Parameter example:*
              ``axes=['domainaxis0', 'time']``
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> c.set_axes('domainaxis1')
    >>> c.has_axes()
    True
    >>> c.get_axes()
    ('domainaxis1',)
    >>> c.del_axes()
    >>> c.has_axes()
    False
    >>> c.get_axes('NO AXES')
    'NO AXES'
    >>> c.del_axes('NO AXES')
    'NO AXES'
    
    >>> c.set_axes(['domainaxis1', 'domainaxis0'])
    
    >>> c.set_axes(['time', 'domainaxis0'])
    
    >>> c.set_axes('time')

        '''
        if copy:
            value = deepcopy(value)
        
        if isinstance(value, basestring):
            value = (value,)
        else:
            value = tuple(value)
            
        return self._set_component('axes', value, copy=False)


    def set_method(self, value, copy=True):
        '''Set the method of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_method`, `get_method`, `has_method`
    
    :Parameters:
    
        value: `str`
            The value for the method.
    
        copy: `bool`, optional
            If True then set a deep copy of *value*.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> c.set_method('minimum')
    >>> c.has_method()
    True
    >>> c.get_method()
    'minimum'
    >>> c.del_method()
    >>> c.has_method()
    False
    >>> c.get_method('NO METHOD')
    'NO METHOD'
    >>> c.del_method('NO METHOD')
    'NO METHOD'

        '''
        return self._set_component('method', value, copy=copy)


    def set_qualifier(self, qualifier, value, copy=True):
        '''Set a qualifier of the cell method.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_qualifier`, `get_qualifier`, `has_qualifier`,
                 `qualifiers`
    
    :Parameters:
    
        qualifier: `str`
            The name of the qualifier to be set.
    
        value: 
            The value for the qualifier.
    
        copy: `bool`, optional
            If True then set a deep copy of *value*.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> c.set_qualifier('where', 'land')
    >>> c.get_qualifier('where')
    'land'

        '''
        if copy:
            value = deepcopy(value)
        
        self._get_component('qualifiers')[qualifier] = value

    
#--- End: class



from . import implementation

from .netcdf import NetCDFWrite


_implementation = implementation()

def write(fields, filename, fmt='NETCDF4', overwrite=True,
          global_attributes=None, variable_attributes=None,
          file_descriptors=None, external=None, Conventions=None,
          datatype=None, least_significant_digit=None,
          endian='native', compress=0, fletcher32=False, shuffle=True,
          verbose=False,
          _implementation=_implementation):
    '''Write field constructs to a netCDF file.

    **File format**
    
    See the *fmt* parameter for details on which output netCDF file
    formats are supported.
    
    **NetCDF variable and dimension names**
    
    These names are stored within constructs read a from dataset, or
    may be set manually. They are used when writing a field construct
    to the file. If a name has not been set then one will be
    constructed (usually based on the standard name if it exists). The
    names may be modified internally to prevent duplication in the
    file.
    
    Each construct, or construct component, that corresponds to a
    netCDF variable has the following methods to get, set and remove a
    netCDF variable name: `!nc_get_variable`, `!nc_set_variable` and
    `!nc_del_variable` method
    
    The domain axis construct has the following methods to get, set
    and remove a netCDF dimension name:
    `~cfdm.DomainAxis.nc_get_dimension`,
    `~cfdm.DomainAxis.nc_set_dimension` and
    `~cfdm.DomainAxis.nc_del_dimension`.
    
    **NetCDF attributes**
    
    Field construct properties may be written as netCDF global
    attributes and/or netCDF data variable attributes. See the
    *file_descriptors*, *global_attributes* and *variable_attributes*
    parameters for details.
    
    **External variables**
    
    Metadata constructs marked as external are omitted from the file
    and referred to via the netCDF "external_variables" global
    attribute. However, omitted constructs may be written to an
    external file (see the *external* parameter for details).
    
    **NetCDF unlimited dimensions**
    
    Domain axis constructs that correspond to NetCDF unlimited
    dimensions may be accessed with the
    `~cfdm.DomainAxis.nc_is_unlimited` and
    `~cfdm.DomainAxis.nc_set_unlimited` methods of a domain axis
    construct.
    
    **NetCDF4 HDF chunk sizes**
    
    HDF5 chunksizes may be set on contruct's data. See the
    `~cfdm.Data.nc_hdf5_chunksizes`,
    `~cfdm.Data.nc_clear_hdf5_chunksizes` and
    `~cfdm.Data.nc_set_hdf5_chunksizes` metods of a `Data` instance.
    
    .. versionadded:: 1.7.0
    
    .. seealso:: `cfdm.read`
    
    :Parameters:
    
        fields: (sequence of) `Field`
            The field constructs to write to the file.
    
        filename: `str`
            The output netCDF file name. Various type of expansion are
            applied to the file names.
    
            Relative paths are allowed, and standard tilde and shell
            parameter expansions are applied to the string.
    
            *Parameter example:*
              The file ``file.nc`` in the user's home directory could
              be described by any of the following:
              ``'$HOME/file.nc'``, ``'${HOME}/file.nc'``,
              ``'~/file.nc'``, ``'~/tmp/../file.nc'``.
      
        fmt: `str`, optional
            The format of the output file. One of:
    
              ==========================  ================================
              *fmt*                       Output file type                
              ==========================  ================================ 
              ``'NETCDF4'``               NetCDF4 format file. This is the   
                                          default.                    
                                                                          
              ``'NETCDF4_CLASSIC'``       NetCDF4 classic format file (see    
                                          below)                     
                                                                          
              ``'NETCDF3_CLASSIC'``       NetCDF3 classic format file 
                                          (limited to file sizes less     
                                          than 2GB).                      
                                                                          
              ``'NETCDF3_64BIT_OFFSET'``  NetCDF3 64-bit offset format
                                          file                            
                                                                          
              ``'NETCDF3_64BIT'``         An alias for
                                          ``'NETCDF3_64BIT_OFFSET'``      
                                                                          
              ``'NETCDF3_64BIT_DATA'``    NetCDF3 64-bit offset format    
                                          file with extensions (see below)      
              ==========================  ================================
    
            By default the format is ``'NETCDF4'``.
    
            All formats support large files (i.e. those greater than
            2GB) except ``'NETCDF3_CLASSIC'``.
    
            ``'NETCDF3_64BIT_DATA'`` is a format that requires version
            4.4.0 or newer of the C library (use `cfdm.environment` to
            see which version if the netCDF-C library is in use). It
            extends the ``'NETCDF3_64BIT_OFFSET'`` binary format to
            allow for unsigned/64 bit integer data types and 64-bit
            dimension sizes.
    
            ``'NETCDF4_CLASSIC'`` files use the version 4 disk format
            (HDF5), but omits features not found in the version 3
            API. They can be read by HDF5 clients. They can also be
            read by netCDF3 clients only if they have been re-linked
            against the netCDF4 library.
    
            ``'NETCDF4'`` files use the version 4 disk format (HDF5)
            and use the new features of the version 4 API.
    
        overwrite: `bool`, optional
            If False then raise an error if the output file
            pre-exists. By default a pre-existing output file is
            overwritten.
    
        Conventions: (sequence of) `str`, optional
             Specify conventions to be recorded by the netCDF global
             "Conventions" attribute. By default the current
             conventions are always included, but if an older CF
             conventions is defined then this is used instead.
    
             *Parameter example:*
               ``Conventions='UGRID-1.0'``
    
             *Parameter example:*
               ``Conventions=['UGRID-1.0']``
    
             *Parameter example:*
               ``Conventions=['CMIP-6.2', 'UGRID-1.0']``
    
             *Parameter example:*
               ``Conventions='CF-1.7'``
    
             *Parameter example:*
               ``Conventions=['CF-1.7', 'UGRID-1.0']``
    
             Note that if the "Conventions" property is set on a field
             construct then it is ignored.
    
        file_descriptors: `dict`, optional
             Create description of file contents netCDF global
             attributes from the specified attributes and their
             values.
    
             If any field construct has a property with the same name
             then it will be written as a netCDF data variable
             attribute, even if it has been specified by the
             *global_attributes* parameter, or has been flagged as
             global on any of the field constructs (see
             `cfdm.Field.nc_global_attributes` for details).
    
             Identification of the conventions being adhered to by the
             file are not specified as a file descriptor, but by the
             *Conventions* parameter instead.
    
             *Parameter example:*
               ``file_attributes={'title': 'my data'}``
    
             *Parameter example:*
               ``file_attributes={'history': 'created 2019-01-01', 'foo': 'bar'}``
    
        global_attributes: (sequence of) `str`, optional
             Create netCDF global attributes from the specified field
             construct properties, rather than netCDF data variable
             attributes.
    
             These attributes are in addition to the following field
             construct properties, which are created as netCDF global
             attributes by default:
             
             * the description of file contents properties (as defined
               by the CF conventions), and
    
             * properties flagged as global on any of the field
               constructs being written (see
               `cfdm.Field.nc_global_attributes` for details).
    
             Note that it is not possible to create a netCDF global
             attribute from a property that has different values for
             different field constructs being written. In this case
             the property will not be written as a netCDF global
             attribute, even if it has been specified by the
             *global_attributes* parameter or is one of the default
             properties, but will appear as an attribute on the netCDF
             data variable corresponding to each field construct that
             contains the property.
    
             Any global attributes that are also specified as file
             descriptors will not be written as netCDF global
             variables, but as netCDF data variable attributes
             instead.
    
             *Parameter example:*
               ``global_attributes='project'``
    
             *Parameter example:*
               ``global_attributes=['project']``
    
             *Parameter example:*
               ``global_attributes=['project', 'experiment']``
    
        variable_attributes: (sequence of) `str`, optional
             Create netCDF data variable attributes from the specified
             field construct properties.
    
             By default, all field construct properties that are not
             created as netCDF global properties are created as
             attributes netCDF data variables. See the
             *global_attributes* parameter for details.
    
             Any field construct property named by the
             *variable_attributes* parameter will always be created as
             a netCDF data variable attribute
    
             *Parameter example:*
               ``variable_attributes='project'``
    
             *Parameter example:*
               ``variable_attributes=['project']``
    
             *Parameter example:*
               ``variable_attributes=['project', 'doi']``
    
        external: `str`, optional  
            Write metadata constructs that have data and are marked as
            external to the named external file. Ignored if there are
            no such constructs.
    
        datatype: `dict`, optional
            Specify data type conversions to be applied prior to
            writing data to disk. This may be useful as a means of
            packing, or because the output format does not support a
            particular data type (for example, netCDF3 classic files
            do not support 64-bit integers). By default, input data
            types are preserved. Any data type conversion is only
            applied to the arrays on disk, and not to the input field
            constructs themselves.
    
            Data types conversions are defined by `numpy.dtype`
            objects in a dictionary whose keys are input data types
            with values of output data types.
    
            *Parameter example:*
              To convert 64-bit integers to 32-bit integers:
              ``datatype={numpy.dtype('int64'):
              numpy.dtype('int32')}``.
           
        endian: `str`, optional
            The endian-ness of the output file. Valid values are
            ``'little'``, ``'big'`` or ``'native'``. By default the
            output is native endian. See the `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for more
            details.
    
            *Parameter example:*
              ``endian='big'``
    
        compress: `int`, optional
            Regulate the speed and efficiency of compression. Must be
            an integer between ``0`` and ``9``. ``0`` means no
            compression; ``1`` is the fastest, but has the lowest
            compression ratio; ``9`` is the slowest but best
            compression ratio. The default value is ``0``. An error is
            raised if compression is requested for a netCDF3 output
            file format. See the `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for more
            details.
    
            *Parameter example:*
              ``compress=4``
        
        least_significant_digit: `int`, optional
            Truncate the input field construct data arrays, but not
            the data arrays of metadata constructs. For a given
            positive integer, N the precision that is retained in the
            compressed data is 10 to the power -N. For example, a
            value of 2 will retain a precision of 0.01. In conjunction
            with the *compress* parameter this produces 'lossy', but
            significantly more efficient, compression. See the
            `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for more
            details.
    
            *Parameter example:*
              ``least_significant_digit=3``
    
        fletcher32: `bool`, optional
            If True then the Fletcher-32 HDF5 checksum algorithm is
            activated to detect compression errors. Ignored if
            *compress* is ``0``. See the `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for details.
    
        shuffle: `bool`, optional
            If True (the default) then the HDF5 shuffle filter (which
            de-interlaces a block of data before compression by
            reordering the bytes by storing the first byte of all of a
            variable's values in the chunk contiguously, followed by
            all the second bytes, and so on) is turned off. By default
            the filter is applied because if the data array values are
            not all wildly different, using the filter can make the
            data more easily compressible.  Ignored if the *compress*
            parameter is ``0`` (which is its default value). See the
            `netCDF4 package
            <http://unidata.github.io/netcdf4-python>`_ for more
            details.
    
        verbose: `bool`, optional
            If True then print a summary of how constructs map to
            output netCDF dimensions, variables and attributes.
    
        _implementation: (subclass of) `CFDMImplementation`, optional
            Define the CF data model implementation that defines field
            and metadata constructs and their components.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> cfdm.write(f, 'file.nc')
    
    >>> cfdm.write(f, 'file.nc', fmt='NETCDF3_CLASSIC')
    
    >>> cfdm.write(f, 'file.nc', external='cell_measures.nc')
    
    >>> cfdm.write(f, 'file.nc', Conventions='CMIP-6.2')

    '''
    # ----------------------------------------------------------------
    # Initialise the netCDF write object
    # ----------------------------------------------------------------
    netcdf = NetCDFWrite(_implementation)

    if fields:
        netcdf.write(fields, filename, fmt=fmt, overwrite=overwrite,
                     global_attributes=global_attributes,
                     variable_attributes=variable_attributes,
                     file_descriptors=file_descriptors,
                     external=external, Conventions=Conventions,
                     datatype=datatype,
                     least_significant_digit=least_significant_digit,
                     endian=endian, compress=compress,
                     shuffle=shuffle, fletcher32=fletcher32,
                     verbose=verbose, extra_write_vars=None)
        



from __future__ import print_function
from builtins import (str, super, zip)

import numpy

from . import mixin
from . import core


class CellMethod(mixin.Container,
                 core.CellMethod):
    '''A cell method construct of the CF data model.

    One or more cell method constructs describe how the cell values of
    the field construct represent the variation of the physical
    quantity within its cells, i.e. the structure of the data at a
    higher resolution.
    
    A single cell method construct consists of a set of axes, a
    "method" property which describes how a value of the field
    construct's data array describes the variation of the quantity
    within a cell over those axes (e.g. a value might represent the
    cell area average), and descriptive qualifiers serving to indicate
    more precisely how the method was applied (e.g. recording the
    spacing of the original data, or the fact that the method was
    applied only over El Nino years).
    
    .. versionadded:: 1.7.0

    '''
    def __str__(self):
        '''Called by the `str` built-in function.

    x.__str__() <==> str(x)
    
    Returns a CF-netCDF-like string of the cell method.
    
    Note that if the intention is to use this string in a CF-netCDF
    cell_methods attribute then, unless they are standard names, the
    axes names will need to be modified to be netCDF dimension names.
    
    .. versionadded:: 1.7.0

        '''     
        string = ['{0}:'.format(axis) for axis in self.get_axes(())]

        string.append(self.get_method(''))

        for portion in ('within', 'where', 'over'):
            q = self.get_qualifier(portion, None)
            if q is not None:
                string.extend((portion, q))
        #--- End: for

        interval = self.get_qualifier('interval', ())
        comment  = self.get_qualifier('comment', None)

        if interval:
            x = ['(']

            y = ['interval: {0}'.format(data) for data in interval]
            x.append(' '.join(y))

            if comment is not None:
                x.append(' comment: {0}'.format(comment))

            x.append(')')

            string.append(''.join(x))

        elif comment is not None:
            string.append('({0})'.format(comment))

        return ' '.join(string)


    def dump(self, display=True, _title=None, _level=0):
        '''A full description of the cell method construct.

    Returns a description the method, all qualifiers and the axes to
    which it applies.
    
    .. versionadded:: 1.7.0
    
    :Parameters:
    
        display: `bool`, optional
            If False then return the description as a string. By
            default the description is printed.
    
    :Returns:
    
        `None` or `str`
            The description. If *display* is True then the description
            is printed and `None` is returned. Otherwise the
            description is returned as a string.

        '''
        indent0 = '    ' * _level

        if _title is None:
            _title = 'Cell Method: '

        return indent0 + _title + str(self)


#    def expand_intervals(self):
#        '''
#        '''
#        c = self.copy()
#
#        n_axes = len(c.get_axes(()))
#        if n_axes > 1:
#            interval = c.get_qualifier('interval', ())            
#            if len(interval) == 1:
#                c.set_property('interval', interval*n_axes)
#
#        return c


#    @classmethod
#    def parse(cls, string, allow_error=False):
#        '''Parse a CF cell_methods string.
#
#:Parameters:
#
#    string: `str`
#        The CF cell_methods string to be parsed into the
#        `cf.CellMethods` object. By default the cell methods will be
#        empty.
#
#    allow_error: `bool`, optional
#
#:Returns:
#
#    `list`
#
#**Examples:**
#
#>>> c = CellMethod.parse('time: minimum within years time: mean over years (ENSO years)')
#>>> print c
#Cell methods    : time: minimum within years
#                  time: mean over years (ENSO years)
#
#>>> c = CellMethod()
#>>> d = c.parse('time: minimum within years time: mean over years (ENSO years)')
#>>> print d
#Cell methods    : time: minimum within years
#                  time: mean over years (ENSO years)
#
#        '''
#        out = []
#        
#        if not string:
#            return out
#
#        # Split the cell_methods string into a list of strings ready
#        # for parsing into the result list. For example,
#        #
#        # 'lat: mean (interval: 1 hour)
#        # 
#        # maps to
#        #
#        # ['lat:', 'mean', '(', 'interval:', '1', 'hour', ')']
#        cell_methods = re_sub('\((?=[^\s])' , '( ', string)
#        cell_methods = re_sub('(?<=[^\s])\)', ' )', cell_methods).split()
#
#        while cell_methods:
#            cm = cls()
#
#            axes  = []
#            while cell_methods:
#                if not cell_methods[0].endswith(':'):
#                    break
#
#                # Check that "name" ebds with colon? How? ('lat: mean (area-weighted) or lat: mean (interval: 1 degree_north comment: area-weighted)')
#
#                axis = cell_methods.pop(0)[:-1]
#
#                axes.append(axis)
#            #--- End: while
#            cm.set_axes(axes)
#
#            if not cell_methods:
#                out.append(cm)
#                break
#
#            # Method
#            cm.set_method(cell_methods.pop(0))
#
#            if not cell_methods:
#                out.append(cm)
#                break
#
#            # Climatological statistics and statistics which apply to
#            # portions of cells
#            while cell_methods[0] in ('within', 'where', 'over'):
#                attr = cell_methods.pop(0)
#                cm.set_property(attr, cell_methods.pop(0))
#                if not cell_methods:
#                    break
#            #--- End: while
#            if not cell_methods: 
#                out.append(cm)
#                break
#
#            # interval and comment
#            interval = []
#            if cell_methods[0].endswith('('):
#                cell_methods.pop(0)
#
#                if not (re_search('^(interval|comment):$', cell_methods[0])):
#                    cell_methods.insert(0, 'comment:')
#                           
#                while not re_search('^\)$', cell_methods[0]):
#                    term = cell_methods.pop(0)[:-1]
#
#                    if term == 'interval':
#                        interval = cell_methods.pop(0)
#                        if cell_methods[0] != ')':
#                            units = cell_methods.pop(0)
#                        else:
#                            units = None
#
#                        try:
#                            parsed_interval = ast_literal_eval(interval)
#                        except:
#                            message = "Cell method interval is incorrectly formatted"
#                            if allow_error:
#                                cm = cls()
#                                cm.set_string(string)
#                                cm.set_error(message)
#                                return [out]
#                            else:
#                                raise ValueError("{}: {}".format(message, string))
#                        #---End: try
#
#                        try:
#                            interval.append(cm._Data(parsed_interval, units=units))
#                        except:
#                            message = "Cell method interval is incorrectly formatted"
#                            if allow_error:
#                                cm = cls()
#                                cm.set_string(string)
#                                cm.set_error(message)
#                                return [cm]
#                            else:
#                                raise ValueError("{}: {}".format(message, string))
#                        #---End: try
#
#                        continue
#                    #--- End: if
#
#                    if term == 'comment':
#                        comment = []
#                        while cell_methods:
#                            if cell_methods[0].endswith(')'):
#                                break
#                            if cell_methods[0].endswith(':'):
#                                break
#                            comment.append(cell_methods.pop(0))
#                        #--- End: while
#                        cm.set_property('comment', ' '.join(comment))
#                    #--- End: if
#
#                #--- End: while 
#
#                if cell_methods[0].endswith(')'):
#                    cell_methods.pop(0)
#            #--- End: if
#
#            n_intervals = len(interval)          
#            if n_intervals > 1 and n_intervals != len(axes):
#                message = "Cell method interval is incorrectly formatted"
#                if allow_error:
#                    cm = cls()
#                    cm.set_string(string)
#                    cm.set_error(message)
#                    return [out]
#                else:
#                    raise ValueError("{}: {}".format(message, string))
#            #---End: if
#
#            if interval:
#                cm.set_property('interval', tuple(interval))
#
#            out.append(cm)
#        #--- End: while
#
#        return out


    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_qualifiers=(), ignore_type=False):
        '''Whether two cell method constructs are the same.

    Equality is strict by default. This means that:
    
    * the descriptive qualifiers must be the same (see the
      *ignore_qualifiers* parameter).
    
    The axes of the cell method constructs are *not* considered,
    because they may only be correctly interpreted by the field
    constructs that contain the cell method constructs in
    question. They are, however, taken into account when two fields
    constructs are tested for equality.
    
    Two real numbers ``x`` and ``y`` are considered equal if
    ``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
    differences) and ``rtol`` (the tolerance on relative differences)
    are positive, typically very small numbers. The data type of the
    numbers is not taken into consideration. See the *atol* and *rtol*
    parameters.
    
    Any type of object may be tested but, in general, equality is only
    possible with another cell method construct, or a subclass of
    one. See the *ignore_type* parameter.
    
    NetCDF elements, such as netCDF variable and dimension names, do
    not constitute part of the CF data model and so are not checked.
    
    .. versionadded:: 1.7.0
    
    :Parameters:
    
        other: 
            The object to compare for equality.
    
        atol: float, optional
            The tolerance on absolute differences between real
            numbers. The default value is set by the `cfdm.ATOL`
            function.
            
        rtol: float, optional
            The tolerance on relative differences between real
            numbers. The default value is set by the `cfdm.RTOL`
            function.
    
        verbose: `bool`, optional
            If True then print information about differences that lead
            to inequality.
    
        ignore_qualifiers: sequence of `str`, optional
            The names of qualifiers to omit from the comparison.
    
        ignore_type: `bool`, optional
            Any type of object may be tested but, in general, equality
            is only possible with another cell method construct, or a
            subclass of one. If *ignore_type* is True then
            ``CellMethod(source=other)`` is tested, rather than the
            ``other`` defined by the *other* parameter.
    
    :Returns: 
      
        `bool`
            Whether the two cell method constructs are equal.
    
    **Examples:**
    
    >>> c.equals(c)
    True
    >>> c.equals(c.copy())
    True
    >>> c.equals('not a cell method')
    False

        '''
        pp = super()._equals_preprocess(other, verbose=verbose,
                                        ignore_type=ignore_type)
        if pp is True or pp is False:
            return pp
        
        other = pp

        # ------------------------------------------------------------
        # Check the methods
        # ------------------------------------------------------------
        if self.get_method(None) != other.get_method(None):
            if verbose:
                print("{0}: Different methods: {1!r} != {2!r}".format(
                    cm0.__class__.__name__, self.get_method(None), other.get_method(None)))
            return False

        # ------------------------------------------------------------
        # Check the qualifiers
        # ------------------------------------------------------------
        self_qualifiers  = self.qualifiers()
        other_qualifiers = other.qualifiers()

#        if ignore_qualifiers:
        for prop in tuple(ignore_qualifiers) + ('interval',):
            self_qualifiers.pop(prop, None)
            other_qualifiers.pop(prop, None)
                
        if set(self_qualifiers) != set(other_qualifiers):
            if verbose:
                for q in set(self_qualifiers).symmetric_difference(other_qualifiers):
                    print("{0}: Non-common qualifier: {1!r}".format( 
                        self.__class__.__name__, q))
            return False

        for qualifier, x in self_qualifiers.items():
            y = other_qualifiers[qualifier]

            if not self._equals(x, y, rtol=rtol, atol=atol,
                                ignore_data_type=True,
                                verbose=verbose):
                if verbose:
                    print("{0}: Different {1} qualifiers: {2!r}, {3!r}".format(
                        self.__class__.__name__, prop, x, y))
                return False
        #--- End: for

        if 'interval' in ignore_qualifiers:
            return True

        intervals0 = self.get_qualifier('interval', ())
        intervals1 = other.get_qualifier('interval', ())
        if intervals0:
            if not intervals1:
                if verbose:
                    print(
                        "{0}: Different interval qualifiers: {1!r} != {2!r}".format(
                            self.__class__.__name__, intervals0, intervals1))
                return False
            #--- End: if
            
            if len(intervals0) != len(intervals1):
                if verbose:
                    print(
                        "{0}: Different numbers of interval qualifiers: {1!r} != {2!r}".format(
                            self.__class__.__name__, intervals0, intervals1))
                    return False
            #--- End: if

            for data0, data1 in zip(intervals0, intervals1):
                if not self._equals(data0, data1,
                                    rtol=rtol, atol=atol,
                                    verbose=verbose,
                                    ignore_data_type=True,
                                    ignore_fill_value=True):
                    if verbose:
                        print(
                            "{0}: Different interval qualifiers: {1!r} != {2!r}".format(
                                self.__class__.__name__, intervals0, intervals1))
                    return False

        elif intervals1:
            if verbose:
                print("{}: Different intervals: {!r} != {!r}".format(
                    self.__class__.__name__, intervals0, intervals1))
            return False
        #--- End: if

        # ------------------------------------------------------------
        # Do NOT check the axes
        # ------------------------------------------------------------

        return True


#    def equivalent(self, other, rtol=None, atol=None, verbose=False):
#        '''True if two cell methods are equivalent, False otherwise.
#
#The `axes` and `interval` attributes are ignored in the comparison.
#
#:Parameters:
#
#    other : 
#        The object to compare for equality.
#
#    atol : float, optional
#        The absolute tolerance for all numerical comparisons, By
#        default the value returned by the `ATOL` function is used.
#
#    rtol : float, optional
#        The relative tolerance for all numerical comparisons, By
#        default the value returned by the `RTOL` function is used.
#
#:Returns: 
#
#    out : bool
#        Whether or not the two instances are equivalent.
#
#**Examples:**
#
#        '''
#        if self is other:
#            return True
#
#        # Check that each instance is the same type
#        if self.__class__ != other.__class__:
#            if verbose:
#                print("{0}: Different types: {0} != {1}".format(
#                    self.__class__.__name__, other.__class__.__name__))
#            return False
#        #--- End: if
#
#        axes0 = self.axes
#        axes1 = other.axes
#            
#        if len(axes0) != len(axes1) or set(axes0) != set(axes1):
#            if verbose:
#                print("{}: Nonequivalent axes: {!r}, {!r}".format(
#                    self.__class__.__name__, axes0, axes1))
#            return False
#        #--- End: if
#
#        argsort = [axes1.index(axis0) for axis0 in axes0]
#        other1 = other.sorted(argsort=argsort)
#        self1 = self
#
#        if not self1.equals(other1, rtol=rtol, atol=atol, ignore=('interval',)):
#            if verbose:
#                print("{0}: Nonequivalent: {1!r}, {2!r}".format(
#                    self.__class__.__name__, self, other))
#            return False
#        #--- End: if
#
#        self_interval  = self1.get_property('interval', ())
#        other_interval = other1.get_property('interval', ())
#        
#        if len(self_interval) != len(other_interval):
#            self1 = self1.expand_intervals(copy=False)
#            other1.expand_intervals(copy=False)
#
#            self_interval  = self1.get_property('interval', ())
#            other_interval = other1.get_property('interval', ())        
#
#            if len(self_interval) != len(other_interval):
#                if verbose:
#                    print(
#"{0}: Different numbers of intervals: {1!r} != {2!r}".format(
#    self.__class__.__name__, self_interval, other_interval))
#                return False
#        #--- End: if
#
##        intervals0 = self1.interval
#        if self_interval:
#            for data0, data1 in zip(self_interval, other_interval):
#                if not data0.allclose(data1, rtol=rtol, atol=atol):
#                    if verbose:
#                        print(
#"{0}: Different interval data: {1!r} != {2!r}".format(
#    self.__class__.__name__, self_interval, other_interval))
#                    return False
#        #--- End: if
#
#        # Still here? Then they are equivalent
#        return True


    def identity(self, default=''):
        '''Return the canonical identity for the cell method construct.

    By default the identity is the first found of the following:
    
    1. The method, preceeded by 'method:'
    2. The value of the *default* parameter.
    
    .. versionadded:: 1.7.0
    
    .. seealso:: `identities`
    
    :Parameters:
    
        default: optional
            If no identity can be found then return the value of the
            default parameter.
       
    :Returns:
    
            The identity.
    
    **Examples:**
    
    >>> c.get_method()
    'minimum;
    >>> c.identity()
    'method:minimum'
    >>> c.identity(default='no identity')
    'method:minimum'
    >>> c.del_method()
    'minimum'
    >>> c.identity()
    ''
    >>> c.identity(default='no identity')
    'no identity'

        '''
        n = self.get_method(None)
        if n is not None:
            return 'method:{0}'.format(n)
            
        return default


    def identities(self):
        '''Return all possible identities.

    The identities comprise:
    
    * The method, preceeded by 'method:'.
    
    .. versionadded:: 1.7.0
    
    .. seealso:: `identity`
    
    :Returns:
    
        `list`
            The identities.
    
    **Examples:**
    
    >>> c.get_method()
    'minimum'
    >>> c.identities()
    ['method:minimum']
    >>> c.del_method()
    'minimum'
    >>> c.identities()
    []

        '''
        out = []

        n = self.get_method(None)
        if n is not None:
            out.append('method:{0}'.format(n))
            
        return out


    def sorted(self, indices=None):
        '''Return a new cell method construct with sorted axes.
        
    The axes are sorted by domain axis construct identifier or
    standard name, and any intervals are sorted accordingly.
    
    .. versionadded:: 1.7.0
    
    :Parameters:
    
        indices: ordered sequence of `int`, optional
            Sort the axes with the given indices. By default the axes
            are sorted by domain axis construct identifier or standard
            name.
    
    :Returns:
    
        `CellMethod`
            A new cell method construct with sorted axes.
    
    **Examples:**
    
    >>> cm = cfdm.CellMethod(axes=['domainaxis1', 'domainaxis0'],
    ...                      method='mean',
    ...                      qualifiers={'interval': [1, 2]})
    >>> cm
    <CellMethod: domainaxis1: domainaxis0: mean (interval: 1 interval: 2)>
    >>> cm.sorted()
    <CellMethod: domainaxis0: domainaxis1: mean (interval: 2 interval: 1)>
    
    >>> cm = cfdm.CellMethod(axes=['domainaxis0', 'area'],
    ...                      method='mean',
    ...                      qualifiers={'interval': [1, 2]})
    >>> cm
    <CellMethod: domainaxis0: area: mean (interval: 1 interval: 2)>
    >>> cm.sorted()
    <CellMethod: area: domainaxis0: mean (interval: 2 interval: 1)>

        '''
        new = self.copy()
        
        axes = new.get_axes(())
        if len(axes) == 1:
            return new

        if indices is None:
            indices = numpy.argsort(axes)
        elif len(indices) != len(axes):
            raise ValueError(
                "Can't sort cell method axes. The given indices ({}) do not correspond to the number of axes ({})".format(
                    indices, axes))

        axes2 = []
        for i in indices:
            axes2.append(axes[i])

        new.set_axes(tuple(axes2))

        intervals = new.get_qualifier('interval', ())
        if len(intervals) <= 1:
            return new

        intervals2 = []
        for i in indices:
            intervals2.append(intervals[i])

        new.set_qualifier('interval', tuple(intervals2))

        return new


#--- End: class



import cfdm, re; q, t = cfdm.read('/home/david/cfdm/docs/_downloads/file.nc')

c = t.cell_methods
x = c.filter_by_type('domain_axis')
xi1 = x.inverse_filter()
xi2 = xi1.inverse_filter()

print c
print x
print xi1
print xi2
print xi2.inverse_filter()



'''cfdm is a complete implementation of the CF data model, that
identifies the fundamental elements of the CF conventions and shows
how they relate to each other, independently of the netCDF encoding.

The central element defined by the CF data model is the field
construct, which corresponds to a CF-netCDF data variable with all of
its metadata.

The cfdm package implements the CF data model for its internal data
structures and so is able to process any CF-compliant dataset. It is
not strict about CF-compliance, however, so that partially conformant
datasets may be modified in memory, as well as ingested from existing
datasets and written to new datasets. This is so that datasets which
are partially conformant may nonetheless be modified in memory and
written to new datasets.

The cfdm package can

    * read field constructs from netCDF datasets,
    * create new field constructs in memory,
    * inspect field constructs,
    * test whether two field constructs are the same,
    * modify field construct metadata and data,
    * create subspaces of field constructs,
    * write field constructs to netCDF datasets on disk,
    * incorporate, and create, metadata stored in external files, and
    * read, write, and create data that have been compressed by
      convention (i.e. ragged or gathered arrays), whilst presenting a
      view of the data in its uncompressed form.

Note that cfdm enables the creation of CF field constructs, but it's
up to the user to use them in a CF-compliant way.

'''

from . import core

__author__       = core.__author__
__date__         = core.__date__
__cf_version__   = core.__cf_version__
__version__      = core.__version__

#_requires = ('numpy',
#             'netCDF4',
#             'cftime',
#             'future')

_requires = ('cftime',)

_error0 = 'cfdm requires the modules {}. '.format(', '.join(_requires))

from distutils.version import LooseVersion
import platform

## Check the version of python
#_minimum_vn = '2.7.0'
#if LooseVersion(platform.python_version()) < LooseVersion(_minimum_vn):
#    raise ValueError(
#        "Bad python version: cfdm requires python version {} or later. Got {}".format(
#            _minimum_vn,  platform.python_version()))
#
#try:
#    import netCDF4
#except ImportError as error1:
#    raise ImportError(_error0+str(error1))
#
#try:
#    import numpy
#except ImportError as error1:
#    raise ImportError(_error0+str(error1))

try:
    import cftime
except ImportError as error1:
    raise ImportError(_error0+str(error1))

#try:
#    import future
#except ImportError as error1:
#    raise ImportError(_error0+str(error1))

## Check the version of netCDF4
#_minimum_vn = '1.4.0'
#if LooseVersion(netCDF4.__version__) < LooseVersion(_minimum_vn):
#    raise ValueError(
#        "Bad netCDF4 version: cfdm requires netCDF4 version {} or later. Got {} at {}".format(
#            minimum_vn, netCDF4.__version__, netCDF4.__file__))
#
## Check the version of numpy
#_minimum_vn = '1.15'
#if LooseVersion(numpy.__version__) < LooseVersion(_minimum_vn):
#    raise ValueError(
#        "Bad numpy version: cfdm requires numpy version {} or later. Got {} at {}".format(
#            minimum_vn, numpy.__version__, numpy.__file__))

from .constants  import masked

from .functions import (CF,
                        environment,
                        ATOL,
                        RTOL)

from .constructs import Constructs

from .data import (Data,
                   Array,
                   CompressedArray,
                   NumpyArray,
                   NetCDFArray,
                   GatheredArray,
                   RaggedContiguousArray,
                   RaggedIndexedArray,
                   RaggedIndexedContiguousArray)

from .count                   import Count
from .index                   import Index
from .list                    import List
from .nodecountproperties     import NodeCountProperties
from .partnodecountproperties import PartNodeCountProperties

from .bounds               import Bounds
from .coordinateconversion import CoordinateConversion
from .datum                import Datum
from .domain               import Domain
from .interiorring         import InteriorRing

from .auxiliarycoordinate import AuxiliaryCoordinate
from .cellmeasure         import CellMeasure
from .cellmethod          import CellMethod
from .coordinatereference import CoordinateReference
from .dimensioncoordinate import DimensionCoordinate
from .domainancillary     import DomainAncillary
from .domainaxis          import DomainAxis
from .field               import Field
from .fieldancillary      import FieldAncillary

from .read_write import (read,
                         write,
                         implementation,
                         CFDMImplementation)




    



from builtins import super

from . import mixin
from . import core


class Count(mixin.NetCDFVariable,
            mixin.NetCDFDimension,
            mixin.NetCDFSampleDimension,
            mixin.PropertiesData,
            core.abstract.PropertiesData):
    '''A count variable required to uncompress a ragged array.

A collection of features stored using a contiguous ragged array
combines all features along a single dimension (the sample dimension)
such that each feature in the collection occupies a contiguous block.

The information needed to uncompress the data is stored in a count
variable that gives the size of each block.

**NetCDF interface**

The netCDF variable name of the count variable may be accessed with
the `nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
`nc_has_variable` methods.

The name of the netCDF dimension spanned by the count variable's data
may be accessed with the `nc_set_dimension`, `nc_get_dimension`,
`nc_del_dimension` and `nc_has_dimension` methods.

The name of the netCDF sample dimension spanned by the compressed data
(that is stored in the "sample_dimension" netCDF attribute and which
does not correspond to a domain axis construct) may be accessed with
the `nc_set_sample_dimension`, `nc_get_sample_dimension`,
`nc_del_sample_dimension` and `nc_has_sample_dimension` methods.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'long_name': 'number of obs for this station'}``

        Properties may also be set after initialisation with the
        `set_properties` and `set_property` methods.

    data: `Data`, optional
        Set the data array. Ignored if the *source* parameter is set.

        The data array may also be set after initialisation with the
        `set_data` method.

    source: optional
        Initialize the properties and data from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)

        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, _key=None, _title=None,
             _create_title=True, _prefix='', _level=0,
             _omit_properties=None, _axes=None, _axis_names=None):
        '''A full description of the count variable.

Returns a description of all properties, including those of
components, and provides selected values of all data arrays.

.. versionadded:: 1.7.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        if _create_title and _title is None: 
            _title = 'Count: ' + self.identity(default='')

        return super().dump(display=display, _key=_key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title, _axes=_axes,
                            _axis_names=_axis_names)
    #--- End: def
    
#--- End: class



from __future__ import print_function

import datetime
import inspect
import os
import unittest

import numpy

import cfdm

class ConstructsTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        f = cfdm.read(self.filename)
        self.assertTrue(len(f)==1, 'f={!r}'.format(f))
        self.f = f[0]

        self.test_only = []
    #--- End: def

    def test_Constructs__repr__str__dump(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        c = f.constructs
        _ = repr(c)
        _ = str(c)
    #--- End: def

    def test_Constructs_key_items_value(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        for key, value in f.constructs.items():
            x = f.constructs.filter_by_key(key)
            self.assertTrue(x.key() == key)
            self.assertTrue(x.value().equals(value))
    #--- End: def


    def test_Constructs_copy_shallow_copy(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        c = self.f.constructs

        d = c.copy()
        self.assertTrue(c.equals(d, verbose=True))
        self.assertTrue(d.equals(c, verbose=True))

        d = c.shallow_copy()
        self.assertTrue(c.equals(d, verbose=True))
        self.assertTrue(d.equals(c, verbose=True))
    #--- End: def

    def test_Constructs_FILTER(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        c = self.f.constructs

        self.assertTrue(len(c) == 20)
        self.assertTrue(len(c.filter_by_identity()) == 20)
#        self.assertTrue(len(c.filter_by_axis())     == 13, str(c.filter_by_axis())) #x
        self.assertTrue(len(c.filter_by_axis())     == 13, str(c.filter_by_axis()))
        self.assertTrue(len(c.filter_by_key())      == 20)
        self.assertTrue(len(c.filter_by_data())     == 13)
        self.assertTrue(len(c.filter_by_property()) == 13, str(c.filter_by_property()))
        self.assertTrue(len(c.filter_by_type())     == 20)
        self.assertTrue(len(c.filter_by_method())   ==  2)
        self.assertTrue(len(c.filter_by_measure())  ==  1)
        self.assertTrue(len(c.filter_by_ncvar())    == 15, str(c.filter_by_ncvar()))
        self.assertTrue(len(c.filter_by_ncdim())    ==  3)

        self.assertTrue(len(c.filter_by_identity('qwerty')) == 0)
        self.assertTrue(len(c.filter_by_key('qwerty'))      == 0)
        self.assertTrue(len(c.filter_by_type('qwerty'))     == 0)
        self.assertTrue(len(c.filter_by_method('qwerty'))   == 0)
        self.assertTrue(len(c.filter_by_measure('qwerty'))  == 0)
        self.assertTrue(len(c.filter_by_ncvar('qwerty'))    == 0)
        self.assertTrue(len(c.filter_by_ncdim('qwerty'))    == 0)
        self.assertTrue(len(c.filter_by_size(-1))           == 0)

        self.assertTrue(len(c.filter_by_identity('latitude'))        == 1)
        self.assertTrue(len(c.filter_by_key('dimensioncoordinate1')) == 1)
        self.assertTrue(len(c.filter_by_type('cell_measure'))        == 1)
        self.assertTrue(len(c.filter_by_method('mean'))              == 1)
        self.assertTrue(len(c.filter_by_measure('area'))             == 1)
        self.assertTrue(len(c.filter_by_ncvar('areacella'))          == 1)
        self.assertTrue(len(c.filter_by_ncdim('grid_longitude'))     == 1)
        self.assertTrue(len(c.filter_by_size(9))                     == 1)
                        
        constructs = c.filter_by_type('auxiliary_coordinate',
                                      'cell_measure',
                                      'cell_method',
                                      'coordinate_reference',
                                      'dimension_coordinate',
                                      'domain_ancillary',
                                      'domain_axis',
                                      'field_ancillary')
        n = 20
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))

        constructs = c.filter_by_type('auxiliary_coordinate')
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.AuxiliaryCoordinate)

        constructs = c.filter_by_type('cell_measure')
        n = 1
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellMeasure)
                        
        constructs = c.filter_by_size(9, 10)
        n = 2
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAxis)

        constructs = c.filter_by_type('cell_method')
        n = 2
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellMethod)

        constructs = c.filter_by_type('dimension_coordinate')
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DimensionCoordinate)

        constructs = c.filter_by_type('coordinate_reference')
        n = 2
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CoordinateReference)

        constructs = c.filter_by_type('domain_ancillary')
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAncillary)

        constructs = c.filter_by_type('field_ancillary')
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.FieldAncillary)

        constructs = c.filter_by_type('domain_axis')
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAxis)

        constructs = c.filter_by_type(*['domain_ancillary'])
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAncillary)

        constructs = c.filter_by_type(*['domain_axis'])
        n =  3
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAxis)

        constructs = c.filter_by_type('domain_ancillary', 'domain_axis')
        n = 6
        self.assertTrue(len(constructs) == n,
                        'Got {} constructs, expected {}'.format(len(constructs), n))

        # Property
        for mode in ([], ['and'], ['or']):
            for kwargs in ({'qwerty': 34},
                           {'standard_name': 'surface_altitude'},
                           {'standard_name': 'surface_altitude', 'units': 'm'},
                           {'standard_name': 'surface_altitude', 'units': 'degrees'},
                           {'standard_name': 'surface_altitude', 'units': 'qwerty'},
            ):                
                d = c.filter_by_property(*mode, **kwargs)
                e = d.inverse_filter()
                self.assertTrue(len(e) == len(c) - len(d))
        
        # Axis
#        for mode in ([], ['and'], ['or'], ['exact'], ['subset'], ['superset']):
        for mode in ('and', 'or', 'exact', 'subset'):
            for args in (['qwerty'],
                         ['domainaxis0'],
                         ['domainaxis0', 'domainaxis1'],
                         ['domainaxis0', 'domainaxis1', 'domainaxis2']):
#                           {'domainaxis0': False},
#                           {'domainaxis0': False, 'domainaxis1': True},
#                           {'domainaxis0': False, 'domainaxis1': True, 'domainaxis2': True},
#                d = c.filter_by_axis(*mode, **kwargs) #x
                d = c.filter_by_axis(mode, *args)
                e = d.inverse_filter()
                self.assertTrue(len(e) == len(c) - len(d))
        #--- End: for
        
        # Inverse filter, filters applied
        self.assertTrue(len(c.filters_applied()) == 0)
        ci = c.inverse_filter()
        self.assertTrue(len(ci) == 0)
        self.assertTrue(len(ci) == len(c) - len(c))

        d = c.filter_by_type('dimension_coordinate', 'auxiliary_coordinate')
        self.assertTrue(len(d.filters_applied()) == 1)
        di = d.inverse_filter()
        self.assertTrue(len(di) == len(c) - len(d))
        
        e = d.filter_by_property(units='degrees')
        self.assertTrue(len(e.filters_applied()) == 2)
        ei = e.inverse_filter(1)
        self.assertTrue(len(e.filters_applied()) == 2)
        self.assertTrue(len(ei) == len(d) - len(e))

        d2 = c.filter_by_type('auxiliary_coordinate')
        e2 = d2.filter_by_naxes(1)
        f2 = e2.inverse_filter(1)
        g2 = f2.inverse_filter(1)
        h2 = g2.inverse_filter(1)
        self.assertTrue(g2.equals(e2, verbose=True))
        self.assertTrue(h2.equals(f2, verbose=True))

        # Unfilter
        self.assertTrue(e.unfilter(1).equals(d, verbose=True))        
        self.assertTrue(e.unfilter(1).unfilter().equals(c, verbose=True))
        self.assertTrue(d.unfilter(1).equals(c, verbose=True))        
        self.assertTrue(c.unfilter(1).equals(c, verbose=True))
    #--- End: def

#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from __future__ import print_function

import datetime
import inspect
import os
import unittest

import numpy

import cfdm

class DomainTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        f = cfdm.read(self.filename)
        self.assertTrue(len(f)==1, 'f={!r}'.format(f))
        self.f = f[0]

        self.test_only = []
    #--- End: def

    def test_Domain__repr__str__dump(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        d = self.f.domain

        _ = repr(d)
        _ = str(d)
        _ = d.dump(display=False)
    #--- End: def

    def test_Domain_equals(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        d = f.domain
        e = d.copy()

        self.assertTrue(d.equals(d, verbose=True))
        self.assertTrue(d.equals(e, verbose=True))
        self.assertTrue(e.equals(d, verbose=True))
    #--- End: def

#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from __future__ import print_function

import datetime
import inspect
import os
import unittest

import numpy

import cfdm

class CellMeasureTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        f = cfdm.read(self.filename)
        self.assertTrue(len(f)==1, 'f={!r}'.format(f))
        self.f = f[0]

        self.test_only = []
    #--- End: def

    def test_CellMeasure__repr__str__dump_construct_type(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        for cm in f.cell_measures.values():
            _ = repr(cm)
            _ = str(cm)
            _ = cm.dump(display=False)
            self.assertTrue(cm.construct_type == 'cell_measure')
    #--- End: def

    def test_CellMeasure(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        cm = f.construct('measure:area')

        self.assertTrue(cm.has_measure())
        self.assertTrue(cm.get_measure() == 'area')
        _ = cm.del_measure()
        self.assertFalse(cm.has_measure())
        self.assertTrue(cm.get_measure(None) == None)
        self.assertTrue(cm.del_measure(None) == None)
        cm.set_measure(_)
        self.assertTrue(cm.has_measure())
        self.assertTrue(cm.get_measure() == 'area')
    #--- End: def

#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)
